# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import errno
import json
import logging
import os
import pickle
from pathlib import Path
import matplotlib
import matplotlib.patches
import matplotlib.pyplot as plt
import matplotlib.widgets
import numpy as np


class TrajectoryPlotter:
    def __init__(self,
                 time_step=None,
                 control_time_step=None,
                 computed_actual_values_time_step=None,
                 pos_limits=None,
                 vel_limits=None,
                 acc_limits=None,
                 jerk_limits=None,
                 torque_limits=None,
                 plot_joint=None,
                 plot_acc_limits=False,
                 plot_time_limits=None,
                 plot_actual_values=False,
                 plot_computed_actual_values=False,
                 plot_actual_torques=False,
                 evaluation_dir=None):

        self._time_step = time_step
        self._control_time_step = control_time_step
        self._computed_actual_values_time_step = computed_actual_values_time_step
        self._plot_acc_limits = plot_acc_limits
        self._plot_time_limits = plot_time_limits  # e.g. [1, 1.5] -> time limits for plotting
        self._plot_actual_values = plot_actual_values
        self._plot_computed_actual_values = plot_computed_actual_values
        self._plot_actual_torques = plot_actual_torques
        self._plot_joint = plot_joint

        self._target_pos = None
        self._trajectory_counter = 0

        self._obstacle_wrapper = None
        self._fig = None
        self._ax = None
        self._ax_indices = None
        self._slider_joint_cartesian = None
        self._ax_slider_link_point = None
        self._joint_cartesian_value = 0
        self._obstacle_index = None
        self._link_index = None
        self._point_index = None

        self._entry_width = 0.05
        self._computed_actual_pos = None
        self._computed_actual_vel = None
        self._computed_actual_acc = None
        self._actual_torque = None

        if time_step:
            self._plot_num_sub_time_steps = int(1000 * time_step)
            self._time_step_counter = None

            self._current_jerk = None
            self._current_acc = None
            self._current_vel = None
            self._current_pos = None

            self._time = None
            self._pos = None
            self._vel = None
            self._acc = None
            self._jerk = None

            self._sub_time = None
            self._sub_pos = None
            self._sub_vel = None
            self._sub_acc = None
            self._sub_jerk = None

            self._pos_limits = pos_limits
            self._vel_limits = vel_limits
            self._acc_limits = acc_limits
            self._jerk_limits = jerk_limits
            self._torque_limits = torque_limits

            self._num_joints = len(self._pos_limits)

            if plot_joint is None:
                self._plot_joint = [True for i in range(self._num_joints)]

            self._episode_counter = 0
            self._zero_vector = [0 for i in range(self._num_joints)]

            self._current_acc_limits = None
            self._actual_pos = None

            self._timestamp = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')

        # matplotlib.use('QT5Agg')
        # select the backend for matplotlib, potential values with GUI support: 'QT5Agg', 'TkAgg'
        matplotlib.rcParams['pdf.fonttype'] = 42
        matplotlib.rcParams['ps.fonttype'] = 42
        matplotlib.rcParams['text.usetex'] = False

        if evaluation_dir is None:
            self._evaluationDir = os.path.join(Path.home(), "safe_motions_evaluation")
        else:
            self._evaluationDir = evaluation_dir

    @property
    def trajectory_time(self):
        return self._time[-1]

    def reset_plotter(self, initial_joint_position):

        self._time_step_counter = 0
        self._trajectory_counter = self._trajectory_counter + 1  # for naming files

        self._episode_counter = self._episode_counter + 1

        self._current_acc = self._zero_vector.copy()
        self._current_vel = self._zero_vector.copy()
        self._current_pos = initial_joint_position.copy()

        self._pos = []
        self._vel = []
        self._acc = []
        self._jerk = []

        self._current_acc_limits = []
        self._current_acc_limits.append([[0, 0] for i in range(self._num_joints)])

        self._sub_pos = []
        self._sub_vel = []
        self._sub_acc = []
        self._sub_jerk = []

        self._actual_pos = []  # not normalized
        self._target_pos = []  # not normalized

        self._computed_actual_pos = []
        self._computed_actual_vel = []
        self._computed_actual_acc = []

        self._actual_torque = []

        self._time = [0]
        self._pos.append([normalize(self._current_pos[i], self._pos_limits[i]) for i in range(len(self._current_pos))])
        self._vel.append([normalize(self._current_vel[i], self._vel_limits[i]) for i in range(len(self._current_vel))])
        self._acc.append([normalize(self._current_acc[i], self._acc_limits[i]) for i in range(len(self._current_acc))])
        self._jerk.append(self._zero_vector.copy())  # zero jerk as initial value

        self._sub_time = [0]
        self._sub_pos.append(self._pos[0].copy())
        self._sub_vel.append(self._vel[0].copy())
        self._sub_acc.append(self._acc[0].copy())
        self._sub_jerk.append(self._jerk[0].copy())

    def display_plot(self, max_time=None, obstacle_wrapper=None, blocking=True):
        if obstacle_wrapper is not None:
            self._obstacle_wrapper = obstacle_wrapper

        num_subplots = 4
        if self._plot_actual_torques:
            num_subplots += 1

        fig, ax = plt.subplots(num_subplots, 1, sharex=True)

        if self._obstacle_wrapper is None or self._obstacle_wrapper.obstacle_scene == 0 \
                or not self._obstacle_wrapper.log_obstacle_data:
            plt.subplots_adjust(left=0.05, bottom=0.04, right=0.95, top=0.98, wspace=0.15, hspace=0.15)
        else:
            plt.subplots_adjust(left=0.05, bottom=0.04, right=0.95, top=0.94, wspace=0.15, hspace=0.15)
            self._add_slider(fig)

        ax_pos = 0
        ax_offset = 1
        ax_vel = 0 + ax_offset
        ax_acc = 1 + ax_offset
        ax_jerk = 2 + ax_offset

        if self._plot_actual_torques:
            ax_torque = 3 + ax_offset
        else:
            ax_torque = None

        self._fig = fig
        self._ax = ax
        self._ax_indices = [ax_pos, ax_vel, ax_acc, ax_jerk, ax_torque]

        if self._joint_cartesian_value == 0:
            self._plot_joint_data(max_time)
        else:
            self._plot_obstacle_data(max_time)

        fig.set_size_inches((24.1, 13.5), forward=False)
        logging.info("Trajectory plotted. Close plot to continue")
        plt.show(block=blocking)

    def _add_slider(self, fig):
        ax_slider_joint_cartesian = fig.add_axes([0.03, 0.96, 0.10, 0.02])
        self._slider_joint_cartesian = ButtonArray(ax_slider_joint_cartesian, label='',
                                                   entries_str=["Joint", "Cartesian"])

        plot_target_points = False
        num_obstacles_and_target_points = self._obstacle_wrapper.num_obstacles
        if plot_target_points:
            num_obstacles_and_target_points + self._obstacle_wrapper.num_target_points
        ax_slider_obstacle = fig.add_axes([0.15, 0.96, self._entry_width * num_obstacles_and_target_points, 0.02])
        entries_str_obstacles = ["Obstacle " + str(i + 1) if self._obstacle_wrapper.obstacle[i].name is None else
                                 self._obstacle_wrapper.obstacle[i].name
                                 for i in range(self._obstacle_wrapper.num_obstacles)]

        if self._obstacle_wrapper.num_target_points > 0 and plot_target_points:
            entries_str_target_points = ["Target " + str(i) if self._obstacle_wrapper.target_point[i].name is None else
                                         self._obstacle_wrapper.target_point[i].name
                                         for i in range(self._obstacle_wrapper.num_target_points)]
            entries_str = entries_str_obstacles + entries_str_target_points
        else:
            entries_str = entries_str_obstacles

        self._slider_obstacle = ButtonArray(ax_slider_obstacle, label='', entries_str=entries_str)

        ax_slider_link = fig.add_axes([0.15 + self._entry_width * num_obstacles_and_target_points + 0.03, 0.96,
                                       len(self._obstacle_wrapper.links_in_use) * self._entry_width, 0.02])
        entries_str = []
        for i in range(len(self._obstacle_wrapper.links_in_use)):
            entries_str.append(self._obstacle_wrapper.link_names[self._obstacle_wrapper.links_in_use[i]])
        self._slider_link = ButtonArray(ax_slider_link, label='', entries_str=entries_str)
        self._slider_link.set_active_values(self._obstacle_wrapper.get_indices_of_observed_links_in_use(
            obstacle_index=0))

        if self._obstacle_index is not None:
            if self._obstacle_index >= num_obstacles_and_target_points:
                self._obstacle_index = num_obstacles_and_target_points - 1
            self._slider_joint_cartesian.set_val(self._joint_cartesian_value)
            self._slider_joint_cartesian.update()
            self._slider_obstacle.set_val(self._obstacle_index)
            self._slider_obstacle.update()
            indices_of_observed_links_in_use = self._obstacle_wrapper.get_indices_of_observed_links_in_use(
                obstacle_index=self._obstacle_index)
            self._slider_link.set_active_values(indices_of_observed_links_in_use)
            slider_link_val = self._obstacle_wrapper.get_index_of_link_in_use(self._obstacle_wrapper.obstacle[
                                                                                  self._obstacle_index].
                                                                              observed_links[self._link_index])
            self._slider_link.set_val(slider_link_val)
            self._slider_link.update()
            self._add_slider_link_point(fig, slider_link_index=slider_link_val, point_index=self._point_index)
            self._slider_link_point.update()

        else:
            self._add_slider_link_point(fig, slider_link_index=self._slider_link.last_val_int)

        self._slider_joint_cartesian.on_changed(self._update_slider_joint_cartesian)
        self._slider_obstacle.on_changed(self._update_slider_obstacle)
        self._slider_link.on_changed(self._update_slider_link)

    def _add_slider_link_point(self, fig, slider_link_index, point_index=0):
        if self._ax_slider_link_point:
            self._ax_slider_link_point.remove()
        link_index = self._obstacle_wrapper.links_in_use[slider_link_index]
        link_point_names = []
        if self._obstacle_wrapper.links[link_index].observe_closest_point:
            link_point_names.append("Closest")
        for i in range(self._obstacle_wrapper.links[link_index].num_observed_points):
            link_point_names.append(self._obstacle_wrapper.links[link_index].observed_points[i].name)
        self._ax_slider_link_point = fig.add_axes([0.15 +
                                                   self._entry_width * (self._obstacle_wrapper.num_obstacles +
                                                                        self._obstacle_wrapper.num_target_points
                                                                        + len(self._obstacle_wrapper.links_in_use))
                                                   + 0.03, 0.96,
                                                   len(link_point_names) * self._entry_width, 0.02])
        self._slider_link_point = ButtonArray(self._ax_slider_link_point, label='', entries_str=link_point_names)
        self._slider_link_point.set_val(point_index)
        self._slider_link_point.on_changed(self._update_slider_link_point)

    def _update_slider_joint_cartesian(self, val):
        int_val = int(self._slider_joint_cartesian.val)
        self._joint_cartesian_value = int_val
        if int_val != self._slider_joint_cartesian.last_val_int:
            if int_val == 0:  # joint
                self._plot_joint_data()
            else:
                self._update_obstacle_indices()
                self._plot_obstacle_data()

    def _update_slider_obstacle(self, val):
        int_val = int(self._slider_obstacle.val)
        if int_val != self._slider_obstacle.last_val_int:
            indices_of_observed_links_in_use = self._obstacle_wrapper.get_indices_of_observed_links_in_use(
                obstacle_index=int_val)
            if self._slider_link.last_val_int not in indices_of_observed_links_in_use:
                self._add_slider_link_point(self._fig, slider_link_index=indices_of_observed_links_in_use[0])
            self._slider_link.set_active_values(indices_of_observed_links_in_use)
            self._update_obstacle_indices(obstacleIndex=int_val)
            if self._slider_joint_cartesian.last_val_int == 1:
                self._plot_obstacle_data()

    def _update_slider_link(self, val):
        int_val = int(self._slider_link.val)
        if int_val != self._slider_link.last_val_int and int_val in self._slider_link.active_values:
            self._add_slider_link_point(self._fig, slider_link_index=int_val)
            self._update_obstacle_indices(linkIndex=int_val)
            if self._slider_joint_cartesian.last_val_int == 1:
                self._plot_obstacle_data()

    def _update_slider_link_point(self, val):
        int_val = int(self._slider_link_point.val)
        if int_val != self._slider_link_point.last_val_int:
            self._update_obstacle_indices(pointIndex=int_val)
            if self._slider_joint_cartesian.last_val_int == 1:
                self._plot_obstacle_data()

    def _clear_axes(self):
        for i in range(len(self._ax)):
            self._ax[i].clear()

    def _update_obstacle_indices(self, **kwargs):
        self._obstacle_index = kwargs.get('obstacle_index', self._slider_obstacle.last_val_int)
        slider_link_value = kwargs.get('link_index', self._slider_link.last_val_int)
        link_val = self._obstacle_wrapper.links_in_use[slider_link_value]

        self._link_index = self._obstacle_wrapper.obstacle[self._obstacle_index].get_link_index(link_val)
        self._point_index = kwargs.get('point_index', self._slider_link_point.last_val_int)

        self._obstacle_wrapper.debug_line_obstacle = self._obstacle_index
        self._obstacle_wrapper.debug_line_link = self._link_index
        self._obstacle_wrapper.debug_line_point = self._point_index

    def _plot_obstacle_data(self, max_time=None):

        self._clear_axes()
        fig = self._fig
        ax = self._ax
        ax_pos = self._ax_indices[0]
        ax_vel = self._ax_indices[1]
        ax_acc = self._ax_indices[2]
        ax_jerk = self._ax_indices[3]
        ax_torque = self._ax_indices[4]

        if ax_torque is not None:
            ax[ax_torque].set_visible(False)
            plt.subplots_adjust(left=0.05, bottom=-0.2, right=0.95, top=0.94, wspace=0.15, hspace=0.15)

        for i in range(len(ax)):
            ax[i].grid(True)
            ax[i].set_xlabel('Time [s]')

        if ax_pos is not None:
            ax[ax_pos].set_ylabel('Distance')

        if ax_vel is not None:
            ax[ax_vel].set_ylabel('Velocity')

        if ax_jerk is not None:
            ax[ax_jerk].set_ylabel('Jerk')

        if ax_acc is not None:
            ax[ax_acc].set_ylabel('Acceleration')

        link_num = self._obstacle_wrapper.obstacle[self._obstacle_index].observed_links[self._link_index]
        point_index = self._point_index
        if self._point_index == 0 and self._obstacle_wrapper.links[link_num].observe_closest_point:
            actual_dis = self._obstacle_wrapper.obstacle[self._obstacle_index].link_data[
                self._link_index].closest_point_distance_actual
            set_dis = self._obstacle_wrapper.obstacle[self._obstacle_index].link_data[
                self._link_index].closest_point_distance_set
        else:
            if self._obstacle_wrapper.links[link_num].observe_closest_point:
                point_index = point_index - 1
            actual_dis = self._obstacle_wrapper.obstacle[self._obstacle_index].link_data[
                self._link_index].observed_point_distance_actual[point_index]
            set_dis = self._obstacle_wrapper.obstacle[self._obstacle_index].link_data[
                self._link_index].observed_point_distance_set[point_index]

        if actual_dis:
            actual_time = np.arange(0, len(actual_dis)) * self._control_time_step
            actual_vel = np.diff(actual_dis) / self._control_time_step
            actual_acc = np.diff(actual_vel) / self._control_time_step
            actual_jerk = np.diff(actual_acc) / self._control_time_step
            if max_time is None or max_time >= actual_time[-1]:
                actual_time_max_index = len(actual_time)
            else:
                actual_time_max_index = np.argmin(np.asarray(actual_time) <= max_time)

            logging.info('Actual Distance: ' + str(np.min(actual_dis)) + ' / ' + str(np.max(actual_dis)) +
                         '; Actual Vel: ' + str(np.min(actual_vel)) + ' / ' + str(np.max(actual_vel)) +
                         '; Actual Acc: ' + str(np.min(actual_acc)) + ' / ' + str(np.max(actual_acc)) +
                         '; Actual Jerk: ' + str(np.min(actual_jerk)) + ' / ' + str(np.max(actual_jerk)))

        if set_dis:
            set_time = np.arange(0, len(set_dis)) * self._control_time_step
            set_vel = np.diff(set_dis) / self._control_time_step
            set_acc = np.diff(set_vel) / self._control_time_step
            set_jerk = np.diff(set_acc) / self._control_time_step
            if max_time is None or max_time >= set_time[-1]:
                set_time_max_index = len(set_time)
            else:
                set_time_max_index = np.argmin(np.asarray(set_time) <= max_time)

            logging.info('Setpoint Distance: ' + str(np.min(set_dis)) + ' / ' + str(np.max(set_dis)) +
                         '; Setpoint Vel: ' + str(np.min(set_vel)) + ' / ' + str(np.max(set_vel)) +
                         '; Setpoint Acc: ' + str(np.min(set_acc)) + ' / ' + str(np.max(set_acc)) +
                         '; Setpoint Jerk: ' + str(np.min(set_jerk)) + ' / ' + str(np.max(set_jerk)))

        line_style = '-'
        marker = '.'
        line_style_actual_value = '--'
        color_actual = 'C0'
        color_set = 'C1'

        timesteps_per_control_step = int(round(self._time_step / self._control_time_step))

        if ax_pos is not None:

            if set_dis:
                ax[ax_pos].plot(set_time[:set_time_max_index], set_dis[:set_time_max_index], color=color_set,
                                linestyle=line_style, label='_nolegend_')
                sub_time = set_time[0:set_time_max_index:timesteps_per_control_step]
                sub_set_dis = set_dis[0:set_time_max_index:timesteps_per_control_step]
                ax[ax_pos].plot(sub_time, sub_set_dis, color=color_set, marker=marker,
                                linestyle='None', label='_nolegend_')

            if actual_dis and self._plot_actual_values:
                ax[ax_pos].plot(actual_time[:actual_time_max_index], actual_dis[:actual_time_max_index],
                                color=color_actual,
                                linestyle=line_style_actual_value, label='_nolegend_')
                sub_time = actual_time[0:actual_time_max_index:timesteps_per_control_step]
                subActualDis = actual_dis[0:actual_time_max_index:timesteps_per_control_step]
                ax[ax_pos].plot(sub_time, subActualDis, color=color_actual, marker=marker,
                                linestyle='None', label='_nolegend_')

        if ax_vel is not None:
            if set_dis:
                vel_time = np.array(
                    set_time[:min(len(set_time) - 1, set_time_max_index)]) + 0.5 * self._control_time_step
                vel_plot = list(set_vel[:min(len(set_time) - 1, set_time_max_index)])
                ax[ax_vel].plot(vel_time, vel_plot, color=color_set,
                                linestyle=line_style, label='_nolegend_')
                sub_time = actual_time[timesteps_per_control_step:set_time_max_index:timesteps_per_control_step]
                sub_set_vel = np.interp(sub_time, vel_time, vel_plot)
                ax[ax_vel].plot(sub_time, sub_set_vel, color=color_set, marker=marker,
                                linestyle='None', label='_nolegend_')

            if actual_dis and self._plot_actual_values:
                vel_time = np.array(
                    actual_time[:min(len(actual_time) - 1, actual_time_max_index)]) + 0.5 * self._control_time_step
                vel_plot = list(actual_vel[:min(len(actual_time) - 1, actual_time_max_index)])
                ax[ax_vel].plot(vel_time, vel_plot, color=color_actual,
                                linestyle=line_style_actual_value, label='_nolegend_')
                sub_time = actual_time[timesteps_per_control_step:actual_time_max_index:timesteps_per_control_step]
                sub_actual_vel = np.interp(sub_time, vel_time, vel_plot)
                ax[ax_vel].plot(sub_time, sub_actual_vel, color=color_actual, marker=marker,
                                linestyle='None', label='_nolegend_')

        if ax_acc is not None:

            if set_dis:
                acc_time = set_time[1:min(len(set_time) - 1, set_time_max_index)]
                acc_plot = list(set_acc[:min(len(set_acc), set_time_max_index - 1)])
                ax[ax_acc].plot(acc_time, acc_plot,
                                color=color_set, linestyle=line_style, label='_nolegend_')
                sub_time = actual_time[timesteps_per_control_step:actual_time_max_index:timesteps_per_control_step]
                sub_set_acc = np.interp(sub_time, acc_time, acc_plot)
                ax[ax_acc].plot(sub_time, sub_set_acc, color=color_set, marker=marker,
                                linestyle='None', label='_nolegend_')

            if actual_dis and self._plot_actual_values:
                acc_time = actual_time[1:min(len(actual_time) - 1, actual_time_max_index)]
                acc_plot = list(actual_acc[:min(len(actual_acc), actual_time_max_index - 1)])
                ax[ax_acc].plot(acc_time, acc_plot,
                                color=color_actual, linestyle=line_style_actual_value, label='_nolegend_')
                sub_time = actual_time[timesteps_per_control_step:actual_time_max_index:timesteps_per_control_step]
                sub_actual_acc = np.interp(sub_time, acc_time, acc_plot)
                ax[ax_acc].plot(sub_time, sub_actual_acc, color=color_actual, marker=marker,
                                linestyle='None', label='_nolegend_')

        if ax_jerk is not None:
            if set_dis:
                jerk_time = np.array(set_time[:min(len(set_time) - 3,
                                                   set_time_max_index - 2)] + 1.5 * self._control_time_step)
                jerk_plot = list(set_jerk[:min(len(set_jerk), set_time_max_index - 2)])
                ax[ax_jerk].plot(jerk_time, jerk_plot,
                                 color=color_set, linestyle=line_style, label='_nolegend_')
                if timesteps_per_control_step == 1:
                    sub_time = set_time[2:set_time_max_index:timesteps_per_control_step]
                else:
                    sub_time = set_time[timesteps_per_control_step:set_time_max_index:timesteps_per_control_step]
                sub_set_jerk = np.interp(sub_time, jerk_time, jerk_plot)
                ax[ax_jerk].plot(sub_time, sub_set_jerk, color=color_set, marker=marker,
                                 linestyle='None', label='_nolegend_')

            if actual_dis and self._plot_actual_values:

                jerk_time = np.array(actual_time[:min(len(actual_time) - 3,
                                                      actual_time_max_index - 2)] + 1.5 * self._control_time_step)
                jerk_plot = list(actual_jerk[:min(len(actual_jerk), actual_time_max_index - 2)])
                ax[ax_jerk].plot(jerk_time, jerk_plot,
                                 color=color_actual, linestyle=line_style_actual_value, label='_nolegend_')
                if timesteps_per_control_step == 1:
                    sub_time = actual_time[2:actual_time_max_index:timesteps_per_control_step]
                else:
                    sub_time = actual_time[timesteps_per_control_step:actual_time_max_index:timesteps_per_control_step]
                sub_actual_jerk = np.interp(sub_time, jerk_time, jerk_plot)
                ax[ax_jerk].plot(sub_time, sub_actual_jerk, color=color_actual, marker=marker,
                                 linestyle='None', label='_nolegend_')

        for i in range(len(ax)):

            if self._plot_time_limits is None:
                ax[i].set_xlim([0, self._time[-1]])
            else:
                ax[i].set_xlim([0, self._plot_time_limits[1] - self._plot_time_limits[0]])

        fig.align_ylabels(ax)

    def _plot_joint_data(self, max_time=None, clear_axes=True, line_style='-'):
        if clear_axes:
            self._clear_axes()
        fig = self._fig
        ax = self._ax
        ax_pos = self._ax_indices[0]
        ax_vel = self._ax_indices[1]
        ax_acc = self._ax_indices[2]
        ax_jerk = self._ax_indices[3]
        ax_torque = self._ax_indices[4]

        if ax_torque is not None:
            if not ax[ax_torque].get_visible():
                ax[ax_torque].set_visible(True)
                plt.subplots_adjust(left=0.05, bottom=0.04, right=0.95, top=0.94, wspace=0.15, hspace=0.15)

        # Positions
        for i in range(len(ax)):
            ax[i].grid(True)
            ax[i].set_xlabel('Time [s]')

        if ax_pos is not None:
            ax[ax_pos].set_ylabel('Position')

        if ax_vel is not None:
            ax[ax_vel].set_ylabel('Velocity')

        if ax_jerk is not None:
            ax[ax_jerk].set_ylabel('Jerk')

        if ax_acc is not None:
            ax[ax_acc].set_ylabel('Acceleration')

        if ax_torque is not None:
            ax[ax_torque].set_ylabel('Torque')

        joint_pos = np.swapaxes(self._pos, 0, 1)  # swap time and joint index dimension for plotting
        joint_vel = np.swapaxes(self._vel, 0, 1)
        joint_acc = np.swapaxes(self._acc, 0, 1)
        joint_jerk = np.swapaxes(self._jerk, 0, 1)

        if self._plot_acc_limits:
            # [time][joint_index][min/max] -> [joint_index][min/max][time]
            joint_acc_limits = np.swapaxes(self._current_acc_limits, 0, 1)
            joint_acc_limits = np.swapaxes(joint_acc_limits, 1, 2)

        joint_sub_pos = np.swapaxes(self._sub_pos, 0, 1)  # swap time and joint index dimension for plotting
        joint_sub_vel = np.swapaxes(self._sub_vel, 0, 1)
        joint_sub_acc = np.swapaxes(self._sub_acc, 0, 1)
        joint_sub_jerk = np.swapaxes(self._sub_jerk, 0, 1)

        if self._actual_pos:
            actual_time = np.arange(0, len(self._actual_pos)) * self._control_time_step
            joint_actual_pos = np.swapaxes(self._actual_pos, 0, 1)

        if self._plot_computed_actual_values and self._computed_actual_pos:
            computed_actual_time = np.arange(0, len(self._computed_actual_pos)) * self._computed_actual_values_time_step
            joint_computed_actual_pos = np.swapaxes(self._computed_actual_pos, 0, 1)
            joint_computed_actual_vel = np.swapaxes(self._computed_actual_vel, 0, 1)
            joint_computed_actual_acc = np.swapaxes(self._computed_actual_acc, 0, 1)

        if self._plot_actual_torques and self._actual_torque:
            actual_time_torque = np.arange(0, len(self._actual_torque)) * self._control_time_step
            joint_actual_torque = np.swapaxes(self._actual_torque, 0, 1)

        marker = '.'
        line_style_actual_value = '--'
        line_style_computed_actual_value = '-.'

        use_computed_actual_value_setup_color = True

        if self._plot_time_limits is not None:
            # shift time so that the plot starts at t=0
            self._time = np.asarray(self._time) - self._plot_time_limits[0]
            self._sub_time = np.asarray(self._sub_time) - self._plot_time_limits[0]
            if self._actual_pos:
                actual_time = actual_time - self._plot_time_limits[0]

        if max_time is None or max_time >= self._time[-1]:
            time_max_index = len(self._time)
        else:
            time_max_index = np.argmin(np.asarray(self._time) <= max_time)
        if max_time is None or max_time >= self._sub_time[-1]:
            sub_time_max_index = len(self._sub_time)
        else:
            sub_time_max_index = np.argmin(np.asarray(self._sub_time) <= max_time)

        if self._actual_pos:
            if max_time is None or max_time >= actual_time[-1]:
                actual_time_max_index = len(actual_time)
            else:
                actual_time_max_index = np.argmin(np.asarray(actual_time) <= max_time)

        if self._plot_actual_torques and self._actual_torque:
            if max_time is None or max_time >= actual_time_torque[-1]:
                actual_time_torque_max_index = len(actual_time_torque)
            else:
                actual_time_torque_max_index = np.argmin(np.asarray(actual_time_torque) <= max_time)

        for j in range(self._num_joints):
            torque_string = ' Actual Torque: ' + str(normalize(np.min(joint_actual_torque[j]),
                                                               self._torque_limits[j])) \
                            + ' / ' + str(normalize(np.max(joint_actual_torque[j]), self._torque_limits[j])) \
                            + ';' if self._plot_actual_torques and self._actual_torque else ""
            logging.info('Joint ' + str(j + 1) + ' (min/max)' + torque_string +
                         ' Jerk: ' + str(np.min(joint_sub_jerk[j])) + ' / ' + str(np.max(joint_sub_jerk[j])) +
                         '; Acc: ' + str(np.min(joint_sub_acc[j])) + ' / ' + str(np.max(joint_sub_acc[j])) +
                         '; Vel: ' + str(np.min(joint_sub_vel[j])) + ' / ' + str(np.max(joint_sub_vel[j])) +
                         '; Pos: ' + str(np.min(joint_sub_pos[j])) + ' / ' + str(np.max(joint_sub_pos[j])))

        for j in range(self._num_joints):
            color = 'C' + str(j)  # "C0", "C1" -> index to the default color cycle
            color_actual = color
            color_limits = color
            color_computed_actual_value_setup = color

            if self._plot_joint[j]:
                label = 'Joint ' + str(j + 1)
                if ax_pos is not None:
                    ax[ax_pos].plot(self._time[:time_max_index], joint_pos[j][:time_max_index], color=color,
                                    marker=marker, linestyle='None', label='_nolegend_')
                    ax[ax_pos].plot(self._sub_time[:sub_time_max_index], joint_sub_pos[j][:sub_time_max_index],
                                    color=color, linestyle=line_style, label=label)

                    if self._actual_pos and self._plot_actual_values:
                        actual_pos_plot = normalize(joint_actual_pos[j], self._pos_limits[j])
                        ax[ax_pos].plot(actual_time[:actual_time_max_index], actual_pos_plot[:actual_time_max_index],
                                        color=color_actual, linestyle=line_style_actual_value, label='_nolegend_')

                    if self._plot_computed_actual_values and self._computed_actual_pos:
                        computed_actual_pos_plot = normalize(joint_computed_actual_pos[j], self._pos_limits[j])
                        computed_actual_value_color = color_computed_actual_value_setup \
                            if use_computed_actual_value_setup_color else color
                        ax[ax_pos].plot(computed_actual_time, computed_actual_pos_plot,
                                        color=computed_actual_value_color,
                                        linestyle=line_style_computed_actual_value, label='_nolegend_')

                if ax_vel is not None:
                    ax[ax_vel].plot(self._time[:time_max_index], joint_vel[j][:time_max_index], color=color,
                                    marker=marker, linestyle='None', label='_nolegend_')
                    ax[ax_vel].plot(self._sub_time[:sub_time_max_index], joint_sub_vel[j][:sub_time_max_index],
                                    color=color, linestyle=line_style, label=label)

                if self._actual_pos and self._plot_actual_values:
                    actual_vel = np.diff(joint_actual_pos[j]) / self._control_time_step
                    actual_vel_plot = normalize(actual_vel, self._vel_limits[j])
                    if ax_vel is not None:
                        ax[ax_vel].plot(np.array(
                            actual_time[:min(len(actual_time) - 1, actual_time_max_index)])
                                        + 0.5 * self._control_time_step,
                                        list(actual_vel_plot[:min(len(actual_time) - 1, actual_time_max_index)]),
                                        color=color_actual, linestyle=line_style_actual_value, label='_nolegend_')

                if self._plot_computed_actual_values and self._computed_actual_pos:
                    computed_actual_vel_plot = normalize(joint_computed_actual_vel[j], self._vel_limits[j])
                    if ax_vel is not None:
                        computed_actual_value_color = color_computed_actual_value_setup \
                            if use_computed_actual_value_setup_color else color
                        ax[ax_vel].plot(computed_actual_time, computed_actual_vel_plot,
                                        color=computed_actual_value_color,
                                        linestyle=line_style_computed_actual_value, label='_nolegend_')

                if ax_acc is not None:
                    ax[ax_acc].plot(self._time[:time_max_index], joint_acc[j][:time_max_index], color=color,
                                    marker=marker, linestyle='None', label='_nolegend_')
                    ax[ax_acc].plot(self._sub_time[:sub_time_max_index], joint_sub_acc[j][:sub_time_max_index],
                                    color=color, linestyle=line_style, label=label)

                if self._actual_pos and self._plot_actual_values:
                    actual_acc = np.diff(actual_vel) / self._control_time_step
                    actual_acc_plot = normalize(actual_acc, self._acc_limits[j])
                    if ax_acc is not None:
                        ax[ax_acc].plot(actual_time[1:min(len(actual_time) - 1, actual_time_max_index)],
                                        list(actual_acc_plot[:min(len(actual_acc_plot), actual_time_max_index - 1)]),
                                        color=color_actual, linestyle=line_style_actual_value, label='_nolegend_')

                if self._plot_computed_actual_values and self._computed_actual_pos:
                    computed_actual_acc_plot = normalize(joint_computed_actual_acc[j], self._acc_limits[j])
                    if ax_acc is not None:
                        computed_actual_value_color = color_computed_actual_value_setup \
                            if use_computed_actual_value_setup_color else color
                        ax[ax_acc].plot(computed_actual_time, computed_actual_acc_plot,
                                        color=computed_actual_value_color, linestyle=line_style_computed_actual_value,
                                        label='_nolegend_')

                if ax_acc is not None:
                    if self._plot_acc_limits and len(joint_acc_limits[0][0]) > 1:
                        for i in range(2):
                            ax[ax_acc].plot(self._time[:time_max_index], joint_acc_limits[j][i][:time_max_index],
                                            color=color_limits, linestyle='--', label='_nolegend_')

                if ax_jerk is not None:
                    ax[ax_jerk].plot(self._time[:time_max_index], joint_jerk[j][:time_max_index], color=color,
                                     marker=marker, linestyle='None', label='_nolegend_')
                    ax[ax_jerk].plot(self._sub_time[:sub_time_max_index], joint_sub_jerk[j][:sub_time_max_index],
                                     color=color, linestyle=line_style, label=label)

                if self._actual_pos and self._plot_actual_values:
                    actual_jerk = np.diff(actual_acc) / self._control_time_step
                    actual_jerk_plot = normalize(actual_jerk, self._jerk_limits[j])
                    if ax_jerk is not None:
                        ax[ax_jerk].plot(np.array(actual_time[:min(len(actual_time) - 3,
                                                                   actual_time_max_index - 2)]
                                                  + 1.5 * self._control_time_step),
                                         list(actual_jerk_plot[:min(len(actual_jerk_plot), actual_time_max_index - 2)]),
                                         color=color_actual, linestyle=line_style_actual_value, label='_nolegend_')

                if self._plot_computed_actual_values and self._computed_actual_pos:
                    computed_actual_jerk = np.diff(joint_computed_actual_acc[j]) / self._control_time_step
                    computed_actual_jerk_plot = normalize(computed_actual_jerk, self._jerk_limits[j])
                    if ax_jerk is not None:
                        computed_actual_value_color = color_computed_actual_value_setup \
                            if use_computed_actual_value_setup_color else color
                        ax[ax_jerk].plot(
                            np.asarray(computed_actual_time[:-1]) + 0.5 * self._computed_actual_values_time_step,
                            computed_actual_jerk_plot, color=computed_actual_value_color,
                            linestyle=line_style_computed_actual_value,
                            label='_nolegend_')

                if ax_torque is not None and self._plot_actual_torques and self._actual_torque:
                    actual_torque_plot = normalize(joint_actual_torque[j], self._torque_limits[j])
                    ax[ax_torque].plot(actual_time_torque[:actual_time_torque_max_index],
                                       actual_torque_plot[:actual_time_torque_max_index], color=color_actual,
                                       linestyle=line_style_actual_value, label=label)

        for i in range(len(ax)):
            ax[i].legend(loc='lower right')
            if self._plot_time_limits is None:
                ax[i].set_xlim([0, self._time[-1]])
            else:
                ax[i].set_xlim([0, self._plot_time_limits[1] - self._plot_time_limits[0]])

        if ax_pos is not None:
            ax[ax_pos].set_ylim([-1.05, 1.05])
        if ax_vel is not None:
            ax[ax_vel].set_ylim([-1.05, 1.05])
        if ax_acc is not None:
            ax[ax_acc].set_ylim([-1.05, 1.05])
        if ax_jerk is not None:
            ax[ax_jerk].set_ylim([-1.05, 1.05])
        if ax_torque is not None:
            ax[ax_torque].set_ylim([-1.05, 1.05])

        fig.align_ylabels(ax)

    def add_data_point(self, normalized_acc, normalized_acc_range=None):
        self._time.append(self._time[-1] + self._time_step)
        last_acc = self._current_acc.copy()
        last_vel = self._current_vel.copy()
        last_pos = self._current_pos.copy()
        self._current_acc = [denormalize(normalized_acc[k], self._acc_limits[k]) for k in
                             range(len(normalized_acc))]
        self._current_jerk = [(self._current_acc[k] - last_acc[k]) / self._time_step
                              for k in range(len(self._current_acc))]
        self._current_vel = [last_vel[k] + 0.5 * self._time_step * (last_acc[k] + self._current_acc[k]) for k in
                             range(len(self._current_vel))]
        self._current_pos = [self._current_pos[k] + last_vel[k] * self._time_step
                             + (1 / 3 * last_acc[k] + 1 / 6 * self._current_acc[k]) * self._time_step ** 2
                             for k in range(len(self._current_pos))]

        self._pos.append([normalize(self._current_pos[k], self._pos_limits[k])
                          for k in range(len(self._current_pos))])
        # normalized position
        self._vel.append([normalize(self._current_vel[k], self._vel_limits[k])
                          for k in range(len(self._current_vel))])
        # normalized velocities
        self._jerk.append([normalize(self._current_jerk[k], self._jerk_limits[k])
                           for k in range(len(self._current_jerk))])
        # normalized jerk
        self._acc.append(normalized_acc.tolist())

        self._current_acc_limits.append(normalized_acc_range)

        # calculate intermediate values for accurate plotting

        for j in range(1, self._plot_num_sub_time_steps + 1):
            t = j / self._plot_num_sub_time_steps * self._time_step
            self._sub_time.append(self._time_step_counter * self._time_step + t)
            self._sub_jerk.append(self._jerk[-1])  # constant jerk as the change in acceleration is linear
            sub_current_acc = [last_acc[k] + ((self._current_acc[k] - last_acc[k]) / self._time_step) * t
                               for k in range(len(self._current_acc))]
            sub_current_vel = [last_vel[k] + last_acc[k] * t +
                               0.5 * ((self._current_acc[k] - last_acc[k]) / self._time_step) * t ** 2
                               for k in range(len(self._current_vel))]
            sub_current_pos = [last_pos[k] + last_vel[k] * t + 0.5 * last_acc[k] * t ** 2 +
                               1 / 6 * ((self._current_acc[k] - last_acc[k]) / self._time_step) * t ** 3
                               for k in range(len(self._current_pos))]

            # normalize acc, vel and pos
            self._sub_acc.append([normalize(sub_current_acc[k], self._acc_limits[k])
                                  for k in range(len(sub_current_acc))])
            self._sub_vel.append([normalize(sub_current_vel[k], self._vel_limits[k])
                                  for k in range(len(sub_current_vel))])
            self._sub_pos.append([normalize(sub_current_pos[k], self._pos_limits[k])
                                  for k in range(len(sub_current_pos))])

        self._time_step_counter = self._time_step_counter + 1

    def add_actual_position(self, actual_position):
        self._actual_pos.append(actual_position)

    def add_actual_torque(self, actual_torque):
        self._actual_torque.append(list(actual_torque))

    def add_computed_actual_value(self, computed_position, computed_velocity, computed_acceleration):
        self._computed_actual_pos.append(computed_position)
        self._computed_actual_vel.append(computed_velocity)
        self._computed_actual_acc.append(computed_acceleration)

    def save_obstacle_data(self, class_name, experiment_name):
        if self._obstacle_wrapper is not None:
            eval_dir = os.path.join(self._evaluationDir, "trajectory_plotter", "logs", class_name,
                                    experiment_name, self._timestamp)

            if not os.path.exists(eval_dir):
                try:
                    os.makedirs(eval_dir)
                except OSError as exc:
                    if exc.errno != errno.EEXIST:
                        raise

            with open(os.path.join(eval_dir, "episode_" + str(self._episode_counter) + "_obstacle_data.pkl"),
                      'wb') as f:
                pickle.dump(self._obstacle_wrapper, f)

    def load_obstacle_data(self, path_to_obstacle_data):
        if os.path.exists(path_to_obstacle_data):
            with open(path_to_obstacle_data, 'rb') as f:
                self._obstacle_wrapper = pickle.load(f)
        else:
            raise

    def save_trajectory(self, class_name, experiment_name):
        trajectory_dict = {}
        trajectory_dict['pos'] = list(self._pos)
        trajectory_dict['vel'] = list(self._vel)
        trajectory_dict['acc'] = list(self._acc)
        trajectory_dict['jerk'] = list(self._jerk)

        trajectory_dict['current_acc_limits'] = list(self._current_acc_limits)

        trajectory_dict['sub_pos'] = list(self._sub_pos)
        trajectory_dict['sub_vel'] = list(self._sub_vel)
        trajectory_dict['sub_acc'] = list(self._sub_acc)
        trajectory_dict['sub_jerk'] = list(self._sub_jerk)

        trajectory_dict['control_time_step'] = float(self._control_time_step)
        trajectory_dict['time_step'] = float(self._time_step)

        trajectory_dict['actual_pos'] = list(self._actual_pos)
        trajectory_dict['actual_torque'] = list(self._actual_torque)

        trajectory_dict['time'] = list(self._time)
        trajectory_dict['sub_time'] = list(self._sub_time)

        trajectory_dict['pos_limits'] = self._pos_limits.tolist()
        trajectory_dict['vel_limits'] = self._vel_limits.tolist()
        trajectory_dict['acc_limits'] = self._acc_limits.tolist()
        trajectory_dict['jerk_limits'] = self._jerk_limits.tolist()
        trajectory_dict['torque_limits'] = self._torque_limits.tolist()

        eval_dir = os.path.join(self._evaluationDir, "trajectory_plotter", "logs", class_name,
                                experiment_name, self._timestamp)

        if not os.path.exists(eval_dir):
            try:
                os.makedirs(eval_dir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

        with open(os.path.join(eval_dir, "episode_" + str(self._episode_counter) + ".json"), 'w') as f:
            f.write(json.dumps(trajectory_dict, default=np_encoder))
            f.flush()

    def load_trajectory(self, path_to_trajectory):

        if os.path.exists(path_to_trajectory):
            with open(path_to_trajectory) as f:
                trajectory_dict = json.load(f)
        else:
            raise

        self._pos = trajectory_dict['pos']
        self._vel = trajectory_dict['vel']
        self._acc = trajectory_dict['acc']
        self._jerk = trajectory_dict['jerk']

        self._current_acc_limits = trajectory_dict['current_acc_limits']

        self._sub_pos = trajectory_dict['sub_pos']
        self._sub_vel = trajectory_dict['sub_vel']
        self._sub_acc = trajectory_dict['sub_acc']
        self._sub_jerk = trajectory_dict['sub_jerk']

        self._control_time_step = trajectory_dict['control_time_step']
        if 'time_step' in trajectory_dict:
            self._time_step = trajectory_dict['time_step']
        self._actual_pos = trajectory_dict['actual_pos']

        if 'actual_torque' in trajectory_dict:
            self._actual_torque = trajectory_dict['actual_torque']

        if 'torque_limits' in trajectory_dict:
            self._torque_limits = trajectory_dict['torque_limits']

        self._time = trajectory_dict['time']
        self._sub_time = trajectory_dict['sub_time']

        self._num_joints = len(self._pos[0])

        self._pos_limits = trajectory_dict['pos_limits']
        self._vel_limits = trajectory_dict['vel_limits']
        self._acc_limits = trajectory_dict['acc_limits']
        self._jerk_limits = trajectory_dict['jerk_limits']


def normalize(value, value_range):
    normalized_value = -1 + 2 * (value - value_range[0]) / (value_range[1] - value_range[0])
    return normalized_value


def denormalize(norm_value, value_range):
    actual_value = value_range[0] + 0.5 * (norm_value + 1) * (value_range[1] - value_range[0])
    return actual_value


def np_encoder(object):
    if isinstance(object, np.generic):
        return object.item()


class ButtonArray(matplotlib.widgets.Slider):

    def __init__(self, ax, label, entries_str, val_init=0, **kwargs):

        self.color_enabled = "white"
        self.color_active = "orange"
        self.color_disabled = "lightgray"
        self.font_size = 10

        self.num_buttons = len(entries_str)

        super(ButtonArray, self).__init__(ax, label, 0, self.num_buttons,
                                          valinit=val_init, valfmt='%1d', **kwargs)

        self._last_val_int = int(val_init)

        self.poly.set_visible(False)
        self.page_buttons = []
        self._active_values = []
        self.vline.set_visible(False)
        for i in range(self.num_buttons):
            button_color = self.color_active if i == val_init else self.color_enabled
            b = matplotlib.patches.Rectangle((float(i) / self.num_buttons, 0), 1. / self.num_buttons, 1,
                                             transform=ax.transAxes, facecolor=button_color)
            ax.add_artist(b)
            self.page_buttons.append(b)
            self._active_values.append(i)
            ax_text = entries_str[i]
            ax.text(float(i) / self.num_buttons + 0.5 / self.num_buttons, 0.45, ax_text,
                    ha="center", va="center", transform=ax.transAxes,
                    fontsize=self.font_size)
        self.valtext.set_visible(False)

    def set_active_values(self, active_values):
        self._active_values = active_values
        self._colorize()

    def update(self):
        self._update(event=None)

    def _update(self, event):
        if event:
            super(ButtonArray, self)._update(event)
        i = int(self.val)
        if i >= self.valmax or i not in self._active_values:
            return
        self._last_val_int = i
        self._colorize()

    def _colorize(self):
        for j in range(self.num_buttons):
            if j in self._active_values:
                self.page_buttons[j].set_facecolor(self.color_enabled)
            else:
                self.page_buttons[j].set_facecolor(self.color_disabled)
        if self._last_val_int not in self._active_values:
            self._last_val_int = self._active_values[0]

        self.page_buttons[self._last_val_int].set_facecolor(self.color_active)

    @property
    def last_val_int(self):
        return self._last_val_int

    @property
    def active_values(self):
        return self._active_values



#!/usr/bin/env python

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np
import argparse
import time
import logging
import os
import sys
import inspect
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))))
from safemotions.envs.safe_motions_env import SafeMotionsEnv


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument('--robot_scene', type=int, default=0)
    parser.add_argument('--obstacle_scene', type=int, default=None)
    parser.add_argument('--use_control_rate_sleep', action='store_true', default=False)
    parser.add_argument('--use_thread_for_control_rate_sleep', action='store_true', default=False)
    parser.add_argument('--obstacle_use_computed_actual_values', action='store_true', default=False)
    parser.add_argument('--log_obstacle_data', action='store_true', default=False)
    parser.add_argument('--plot_trajectory', action='store_true', default=False)
    parser.add_argument('--plot_actual_values', action='store_true', default=False)
    parser.add_argument('--pos_limit_factor', type=float, default=1.0)
    parser.add_argument('--vel_limit_factor', type=float, default=1.0)
    parser.add_argument('--acc_limit_factor', type=float, default=1.0)
    parser.add_argument('--jerk_limit_factor', type=float, default=1.0)
    parser.add_argument('--torque_limit_factor', type=float, default=1.0)

    args = parser.parse_args()

    seed = None  # None or an integer (for debugging purposes)
    if seed is not None:
        np.random.seed(seed)
    experiment_name = "random_agent"

    use_real_robot = False
    real_robot_debug_mode = False
    # if true, the joint commands are not send to the real robot but all other computations stay the same.
    use_gui = True
    control_time_step = None
    # None: use default; if provided, the trajectory timestep must be a multiple of the control_time_step
    use_control_rate_sleep = args.use_control_rate_sleep
    # whether to slow down the simulation such that it is synchronized with the control rate;
    # not required if the simulation is slower anyway
    use_thread_for_control_rate_sleep = args.use_thread_for_control_rate_sleep
    # if true, the control rate is more accurate but actual values have to be computed in advance
    obstacle_use_computed_actual_values = args.obstacle_use_computed_actual_values
    # use computed instead of measured actual values to update the obstacle wrapper -> can be computed in advance
    # -> required to use a separate thread for real-time execution

    render_video = False
    # whether to render a video
    camera_angle = 1
    # camera_angle for video rendering; predefined settings are defined in video_recording.py

    pos_limit_factor = args.pos_limit_factor
    vel_limit_factor = args.vel_limit_factor
    acc_limit_factor = args.acc_limit_factor
    jerk_limit_factor = args.jerk_limit_factor
    torque_limit_factor = args.torque_limit_factor
    # the maximum values for the position, velocity, acceleration, jerk and torque for each joint are multiplied
    # with the factors defined above

    robot_scene = args.robot_scene
    # 0: one iiwa robot, 1: two iiwa robots, 2: three iiwa robots

    if robot_scene == 0:
        num_robots = 1
    if robot_scene == 1:
        num_robots = 2
    if robot_scene == 2:
        num_robots = 3

    plot_trajectory = args.plot_trajectory
    save_trajectory_plot = False
    plot_joint = [True, True, True, True, True, True, True] * num_robots
    # selects the joints to be plotted;
    plot_acc_limits = False
    plot_actual_values = args.plot_actual_values
    plot_computed_actual_values = False
    plot_actual_torques = True
    plot_time_limits = None
    log_obstacle_data = args.log_obstacle_data
    # plots of obstacle date are only available if log_obstacle_data is True
    save_obstacle_data = False
    # store the data of the obstacle wrapper as a pickle object; saves the trajectory plot as well
    store_actions = False
    # save the list of predicted actions

    acc_limit_factor_braking = 0.75  # for brake trajectories, relative to acc_limit_factor
    jerk_limit_factor_braking = 0.75  # for brake trajectories, relative to the corresponding maximum jerk

    position_controller_time_constants = [0.030, 0.030, 0.030, 0.030, 0.030, 0.030, 0.030] * num_robots
    # time constants in seconds given per joint; calculate the expected actual values by
    # modelling the behaviour of the trajectory controller without reading actual values from sensor data

    online_trajectory_duration = 8  # time in seconds
    online_trajectory_time_step = 0.1  # in seconds

    if args.obstacle_scene is None:
        # 0: no obstacles, 1: table only; 2: Table + walls, 3: table + walls + monitor (no pivot),
        # 4: table + walls + monitor (pivot)
        if robot_scene == 0:
            obstacle_scene = 3
        else:
            obstacle_scene = 1
    else:
        obstacle_scene = args.obstacle_scene

    use_braking_trajectory_method = True
    # collision avoidance by calculating an alternative safe behavior
    collision_check_time = 0.05
    # time in seconds between collision checks; the trajectory timestep should be a multiple of this time
    check_braking_trajectory_closest_points = True
    # True: Closed points are checked (including self-collision if activated);
    closest_point_safety_distance = 0.05
    # minimum distance that should be guaranteed by the collision avoidance method for closest points
    check_braking_trajectory_observed_points = False
    # True: Observed points are checked;
    observed_point_safety_distance = 0.05
    # minimum distance that should be guaranteed by the collision avoidance method for observed points
    observed_link_point_scene = 0
    # 0: no observed link points, 1: observe link 6 and 7 with a single sphere,
    # 2: observe the robot body starting from link 3 with in total 6 spheres
    visualize_bounding_spheres = False
    # whether to visualize the bounding spheres of observed link points, requires log_obstacle_data to be True
    check_braking_trajectory_torque_limits = True
    # torque constraint adherence by checking the actual torque of an alternative safe behavior

    use_target_points = True  # activate a reaching task with random target points
    target_link_name = "iiwa_link_7"  # name of the target link for target point reaching
    target_link_offset = [0, 0, 0.126]  # relative offset between the frame of the target link and the target link point
    target_point_cartesian_range_scene = num_robots - 1
    # different settings for the cartesian range of the target point as defined in obstacle_torque_prevention.py
    # 0: Cartesian range used for a single robot, 1: Cartesian range used for two robots,
    # 2: Cartesian range used for three robots
    target_point_radius = 0.065
    # a target point is considered as reached if the distance to the target point is smaller than the specified radius
    target_point_sequence = 1  # 0: target points for all robots (T_S), 1: alternating target points (T_A)
    target_point_reached_reward_bonus = 5  # reward bonus if a target point is reached
    target_point_max_jerk_punishment = 1.0  # the jerk punishment is multiplied by this factor
    target_point_reward_factor = 1.0  # the target point reward is multiplied by this factor

    if use_target_points:
        obs_add_target_point_pos = True  # add the Cartesian position of the target point to the observation
        obs_add_target_point_relative_pos = True
        # add the difference between the target point and the target link point in Cartesian space (for each coordinate)
    else:
        obs_add_target_point_pos = False
        obs_add_target_point_relative_pos = False

    punish_action = True  # optional action penalty
    action_punishment_min_threshold = 0.9  # (joint) actions greater than this threshold are punished
    action_max_punishment = 1  # maximum punishment if the absolute value of the highest (joint) action is 1

    punish_braking_trajectory_min_distance = False
    # whether to add a penalty based on the minimum distance between the braking trajectory and
    # an observed link or obstacle
    braking_trajectory_min_distance_max_threshold = 0.05
    braking_trajectory_min_distance_max_punishment = 1

    punish_braking_trajectory_max_torque = False
    # whether to add a penalty based on the maximum torque during the simulation of the braking trajectory
    braking_trajectory_max_torque_min_threshold = 0.8
    braking_trajectory_max_torque_max_punishment = 0.5

    random_agent = True  # whether to use random actions

    env_config = dict(env="ObstacleTorqueEnv", experiment_name=experiment_name,
                      use_gui=use_gui, use_control_rate_sleep=use_control_rate_sleep,
                      use_thread_for_control_rate_sleep=use_thread_for_control_rate_sleep,
                      control_time_step=control_time_step,
                      render_video=render_video, camera_angle=camera_angle,
                      use_real_robot=use_real_robot,
                      real_robot_debug_mode=real_robot_debug_mode,
                      pos_limit_factor=pos_limit_factor,
                      vel_limit_factor=vel_limit_factor,
                      acc_limit_factor=acc_limit_factor,
                      jerk_limit_factor=jerk_limit_factor,
                      torque_limit_factor=torque_limit_factor,
                      plot_trajectory=plot_trajectory, save_trajectory_plot=save_trajectory_plot,
                      plot_joint=plot_joint, plot_acc_limits=plot_acc_limits,
                      plot_actual_values=plot_actual_values,
                      plot_computed_actual_values=plot_computed_actual_values,
                      plot_actual_torques=plot_actual_torques,
                      plot_time_limits=plot_time_limits,
                      log_obstacle_data=log_obstacle_data,
                      save_obstacle_data=save_obstacle_data,
                      store_actions=store_actions,
                      acc_limit_factor_braking=acc_limit_factor_braking,
                      jerk_limit_factor_braking=jerk_limit_factor_braking,
                      online_trajectory_duration=online_trajectory_duration,
                      online_trajectory_time_step=online_trajectory_time_step,
                      position_controller_time_constants=position_controller_time_constants,
                      robot_scene=robot_scene,
                      obstacle_scene=obstacle_scene,
                      observed_link_point_scene=observed_link_point_scene,
                      visualize_bounding_spheres=visualize_bounding_spheres,
                      obstacle_use_computed_actual_values=obstacle_use_computed_actual_values,
                      use_braking_trajectory_method=use_braking_trajectory_method,
                      collision_check_time=collision_check_time,
                      check_braking_trajectory_observed_points=check_braking_trajectory_observed_points,
                      check_braking_trajectory_closest_points=check_braking_trajectory_closest_points,
                      check_braking_trajectory_torque_limits=check_braking_trajectory_torque_limits,
                      closest_point_safety_distance=closest_point_safety_distance,
                      observed_point_safety_distance=observed_point_safety_distance,
                      use_target_points=use_target_points,
                      target_point_cartesian_range_scene=target_point_cartesian_range_scene,
                      obs_add_target_point_pos=obs_add_target_point_pos,
                      obs_add_target_point_relative_pos=obs_add_target_point_relative_pos,
                      punish_action=punish_action,
                      action_punishment_min_threshold=action_punishment_min_threshold,
                      action_max_punishment=action_max_punishment,
                      target_link_name=target_link_name,
                      target_link_offset=target_link_offset,
                      punish_braking_trajectory_min_distance=punish_braking_trajectory_min_distance,
                      braking_trajectory_min_distance_max_threshold=braking_trajectory_min_distance_max_threshold,
                      braking_trajectory_min_distance_max_punishment=braking_trajectory_min_distance_max_punishment,
                      punish_braking_trajectory_max_torque=punish_braking_trajectory_max_torque,
                      braking_trajectory_max_torque_min_threshold=braking_trajectory_max_torque_min_threshold,
                      braking_trajectory_max_torque_max_punishment=braking_trajectory_max_torque_max_punishment,
                      target_point_radius=target_point_radius,
                      target_point_sequence=target_point_sequence,
                      target_point_reached_reward_bonus=target_point_reached_reward_bonus,
                      target_point_max_jerk_punishment=target_point_max_jerk_punishment,
                      target_point_reward_factor=target_point_reward_factor,
                      seed=seed, random_agent=random_agent)

    env = SafeMotionsEnv(**env_config)
    num_episodes = 20

    for _ in range(num_episodes):
        done = False
        step = 0
        env.reset()
        sum_of_rewards = 0
        start = time.time()

        while not done:
            obs, reward, done, info = env.step(action=None)
            step += 1
            sum_of_rewards += reward

        end = time.time()
        logging.info("Last episode took %s seconds. Trajectory duration: %s seconds.", end - start,
                     step * online_trajectory_time_step)
        logging.info("Reward: %s", sum_of_rewards)

    env.close()



#!/usr/bin/env python

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import argparse
import json
import os
import ray
from ray import tune
import multiprocessing
from collections import defaultdict
import sys
import logging
import inspect
current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.append(os.path.dirname(current_dir))
from safemotions.envs.safe_motions_env import SafeMotionsEnv

METRIC_OPS = ['sum', 'average', 'max', 'min']

# Termination reason
TERMINATION_UNSET = -1
TERMINATION_SUCCESS = 0
TERMINATION_JOINT_LIMITS = 1
TERMINATION_TRAJECTORY_LENGTH = 2


def on_episode_start(info):
    episode = info['episode']
    for op in METRIC_OPS:
        episode.user_data[op] = defaultdict(list)


def on_episode_step(info):
    episode = info['episode']
    episode_info = episode.last_info_for()
    if episode_info:
        for op in list(episode_info.keys() & METRIC_OPS):
            for k, v in episode_info[op].items():
                episode.user_data[op][k].append(v)


def on_episode_end(info):
    def __apply_op_on_list(operator, data_list):
        if operator == 'sum':
            return sum(data_list)
        elif operator == 'average':
            return sum(data_list) / len(data_list)
        elif operator == 'max':
            return max(data_list)
        elif operator == 'min':
            return min(data_list)

    episode = info['episode']
    episode_info = episode.last_info_for()
    episode_length = episode_info['episode_length']
    trajectory_length = episode_info['trajectory_length']

    for op in METRIC_OPS:
        for k, v in episode.user_data[op].items():
            episode.custom_metrics[k + "_" + op] = __apply_op_on_list(op, episode.user_data[op][k])

    for k, v in episode_info.items():
        if k.startswith("obstacles"):
            episode.custom_metrics[k] = v

    episode.custom_metrics['episode_length'] = float(episode_length)
    episode.custom_metrics['trajectory_length'] = trajectory_length

    if episode_info['termination_reason'] == TERMINATION_SUCCESS:
        episode.custom_metrics['success_rate'] = 1.0
    else:
        episode.custom_metrics['success_rate'] = 0.0

    if episode_info['termination_reason'] == TERMINATION_JOINT_LIMITS:
        episode.custom_metrics['joint_limit_violation_termination_rate'] = 1.0
    else:
        episode.custom_metrics['joint_limit_violation_termination_rate'] = 0.0


def on_train_result(info):
    info['result']['callback_ok'] = True


callbacks = {'on_episode_start': on_episode_start,
             'on_episode_step': on_episode_step,
             'on_episode_end': on_episode_end,
             'on_train_result': on_train_result}


def _make_env_config():

    env_config = {
        'experiment_name': args.name,
        'm_prev': args.m_prev,
        'pos_limit_factor': args.pos_limit_factor,
        'vel_limit_factor': args.vel_limit_factor,
        'acc_limit_factor': args.acc_limit_factor,
        'jerk_limit_factor': args.jerk_limit_factor,
        'torque_limit_factor': args.torque_limit_factor,
        'normalize_reward_to_frequency': args.normalize_reward_to_frequency,
        'online_trajectory_duration': args.online_trajectory_duration,
        'online_trajectory_time_step': args.online_trajectory_time_step,
        'obs_add_target_point_pos': args.obs_add_target_point_pos,
        'obs_add_target_point_relative_pos': args.obs_add_target_point_relative_pos,
        'punish_action': args.punish_action,
        'action_punishment_min_threshold': args.action_punishment_min_threshold,
        'action_max_punishment': args.action_max_punishment,
        'obstacle_scene': args.obstacle_scene,
        'log_obstacle_data': False,
        'use_braking_trajectory_method': args.use_braking_trajectory_method,
        'check_braking_trajectory_torque_limits': args.check_braking_trajectory_torque_limits,
        'closest_point_safety_distance': args.closest_point_safety_distance,
        'use_target_points': args.use_target_points,
        'acc_limit_factor_braking': args.acc_limit_factor_braking,
        'jerk_limit_factor_braking': args.jerk_limit_factor_braking,
        'punish_braking_trajectory_min_distance': args.punish_braking_trajectory_min_distance,
        'braking_trajectory_min_distance_max_threshold': args.braking_trajectory_min_distance_max_threshold,
        'braking_trajectory_min_distance_max_punishment': args.braking_trajectory_min_distance_max_punishment,
        'punish_braking_trajectory_max_torque': args.punish_braking_trajectory_max_torque,
        'braking_trajectory_max_torque_min_threshold': args.braking_trajectory_max_torque_min_threshold,
        'braking_trajectory_max_torque_max_punishment': args.braking_trajectory_max_torque_max_punishment,
        'robot_scene': args.robot_scene,
        'target_point_cartesian_range_scene': args.target_point_cartesian_range_scene,
        'target_point_radius': args.target_point_radius,
        'target_point_sequence': args.target_point_sequence,
        'target_point_reached_reward_bonus': args.target_point_reached_reward_bonus,
        'target_point_reward_factor': args.target_point_reward_factor,
        'target_link_offset': args.target_link_offset,
        'obstacle_use_computed_actual_values': args.obstacle_use_computed_actual_values
    }

    return env_config


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    config = {
        'model': {
            'conv_filters': None,
            'fcnet_hiddens': [256, 128],
            'fcnet_activation': "selu",
            'use_lstm': False,
        },
        'gamma': 0.99,
        'use_gae': True,
        'lambda': 1.0,
        'kl_coeff': 0.2,
        'sample_batch_size': 4096,
        'train_batch_size': 16384,
        'sgd_minibatch_size': 1024,
        'num_sgd_iter': 16,
        'lr': 5e-5,
        'lr_schedule': None,
        'vf_loss_coeff': 1.0,
        'entropy_coeff': 0.00,
        'clip_param': 0.3,
        'vf_clip_param': 10.0,
        'kl_target': 0.01,
        'batch_mode': 'complete_episodes',
    }

    algorithm = "PPO"
    parser = argparse.ArgumentParser()
    parser.add_argument('--name', type=str, default="default_name")
    parser.add_argument('--logdir', type=str, default=None)
    parser.add_argument('--checkpoint', type=str, default=None,
                        help="Path to a checkpoint if training should be continued.")
    parser.add_argument('--time', type=int, required=True,
                        help="Total time of the training in hours.")
    parser.add_argument('--iterations_per_checkpoint', type=int, default=500,
                        help="The number of training iterations per checkpoint")
    parser.add_argument('--num_workers', type=int, default=None)
    parser.add_argument('--num_gpus', type=int, default=None)
    parser.add_argument('--m-prev', type=int, default=1)
    parser.add_argument('--pos_limit_factor', type=float, default=1.0)
    parser.add_argument('--vel_limit_factor', type=float, default=1.0)
    parser.add_argument('--acc_limit_factor', type=float, default=1.0)
    parser.add_argument('--jerk_limit_factor', type=float, default=1.0)
    parser.add_argument('--torque_limit_factor', type=float, default=1.0)
    parser.add_argument('--normalize_reward_to_frequency', dest='normalize_reward_to_frequency', action='store_true',
                        default=False)
    parser.add_argument('--batch_size_factor', type=float, default=1.0)
    parser.add_argument('--online_trajectory_duration', type=float, default=8.0)
    parser.add_argument('--online_trajectory_time_step', type=float, default=0.1)
    parser.add_argument('--obs_add_target_point_pos', action='store_true', default=False)
    parser.add_argument('--obs_add_target_point_relative_pos', action='store_true', default=False)
    parser.add_argument('--punish_action', dest='punish_action', action='store_true', default=False)
    parser.add_argument('--action_punishment_min_threshold', type=float, default=0.9)
    parser.add_argument('--action_max_punishment', type=float, default=0.5)
    parser.add_argument('--punish_braking_trajectory_min_distance', action='store_true', default=False)
    parser.add_argument('--braking_trajectory_min_distance_max_threshold', type=float, default=0.05)
    parser.add_argument('--braking_trajectory_min_distance_max_punishment', type=float, default=1.0)
    parser.add_argument('--punish_braking_trajectory_max_torque', action='store_true', default=False)
    parser.add_argument('--braking_trajectory_max_torque_min_threshold', type=float, default=0.8)
    parser.add_argument('--braking_trajectory_max_torque_max_punishment', type=float, default=0.5)
    parser.add_argument('--obstacle_scene', type=int, default=0)
    parser.add_argument('--use_braking_trajectory_method', action='store_true', default=False)
    parser.add_argument('--check_braking_trajectory_torque_limits', action='store_true', default=False)
    parser.add_argument('--closest_point_safety_distance', type=float, default=0.1)
    parser.add_argument('--use_target_points', action='store_true', default=False)
    parser.add_argument('--acc_limit_factor_braking', type=float, default=0.75)
    parser.add_argument('--jerk_limit_factor_braking', type=float, default=0.75)
    parser.add_argument('--robot_scene', type=int, default=0)
    parser.add_argument('--target_point_cartesian_range_scene', type=int, default=0)
    parser.add_argument('--target_point_radius', type=float, default=0.065)
    parser.add_argument('--target_point_sequence', type=int, default=0)
    parser.add_argument('--target_point_reached_reward_bonus', type=float, default=0.00)
    parser.add_argument('--target_link_offset', type=json.loads, default="[0, 0, 0.126]")
    parser.add_argument('--target_point_reward_factor', type=float, default=1.0)
    parser.add_argument('--obstacle_use_computed_actual_values', action='store_true', default=False)
    args = parser.parse_args()
    env_name = "SafeMotionsEnv"
    tune.register_env(env_name, lambda config_args: SafeMotionsEnv(**config_args))
    config.update(env=env_name)

    if args.checkpoint is not None:
        if not os.path.isdir(args.checkpoint) and not os.path.isfile(args.checkpoint):
            checkpoint_path = os.path.join(current_dir, "trained_networks", args.checkpoint)
        else:
            checkpoint_path = args.checkpoint

        if os.path.isdir(checkpoint_path):
            if os.path.basename(checkpoint_path) == "checkpoint":
                checkpoint_path = os.path.join(checkpoint_path, "checkpoint")
            else:
                checkpoint_path = os.path.join(checkpoint_path, "checkpoint", "checkpoint")

        if not os.path.isfile(checkpoint_path):
            raise ValueError("Could not find checkpoint {}".format(checkpoint_path))

        params_dir = os.path.dirname(os.path.dirname(checkpoint_path))
        params_path = os.path.join(params_dir, "params.json")

        with open(params_path) as params_file:
            checkpoint_config = json.load(params_file)
        config['env_config'] = checkpoint_config['env_config']
        config['sample_batch_size'] = checkpoint_config['sample_batch_size']
        config['train_batch_size'] = checkpoint_config['train_batch_size']
        config['sgd_minibatch_size'] = checkpoint_config['sgd_minibatch_size']

    else:
        checkpoint_path = None
        config['env_config'] = _make_env_config()
        config['sample_batch_size'] = int(config['sample_batch_size'] * args.batch_size_factor)
        config['train_batch_size'] = int(config['train_batch_size'] * args.batch_size_factor)
        config['sgd_minibatch_size'] = int(config['sgd_minibatch_size'] * args.batch_size_factor)

    if args.logdir is None:
        experiment_path = config['env_config']['experiment_name']
    else:
        experiment_path = os.path.join(args.logdir, config['env_config']['experiment_name'])

    ray.init()
    config['callbacks'] = callbacks

    if args.num_workers is None:
        config['num_workers'] = int(multiprocessing.cpu_count() * 0.75)
    else:
        config['num_workers'] = args.num_workers

    if args.num_gpus is not None:
        config['num_gpus'] = args.num_gpus

    stop = {'time_total_s': args.time * 3600}

    experiment = {
        experiment_path: {
            'run': algorithm,
            'env': env_name,
            'stop': stop,
            'config': config,
            'checkpoint_freq': args.iterations_per_checkpoint,
            'checkpoint_at_end': True,
            'keep_checkpoints_num': 10,
            'max_failures': 0,
            'restore': checkpoint_path
        }
    }

    tune.run_experiments(experiment)



# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import pybullet as p
import datetime
from abc import ABC
from gym.wrappers.monitoring.video_recorder import VideoRecorder
from safemotions.envs.safe_motions_base import SafeMotionsBase

RENDER_MODES = ["human", "rgb_array"]
VIDEO_FRAME_RATE = 60
VIDEO_WIDTH = 1920
VIDEO_HEIGHT = 1080


def compute_view_matrix(camera_angle=0):
    if camera_angle == 0:
        cam_target_pos = (0, 0, 0)
        cam_dist = 1.75
        yaw = 90
        pitch = -70
        roll = 0

    elif camera_angle == 1:
        cam_target_pos = (-0.25, 0, 0)
        cam_dist = 1.95
        yaw = 90
        pitch = -40
        roll = 0

    elif camera_angle == 2:
        yaw = 59.59992599487305
        pitch = -49.400054931640625
        cam_dist = 2.000002861022949
        cam_target_pos = (0.0, 0.0, 0.0)
        roll = 0

    elif camera_angle == 3:
        yaw = 64.39994049072266
        pitch = -37.000003814697266
        cam_dist = 2.000002861022949
        cam_target_pos = (0.0, 0.0, 0.0)
        roll = 0

    elif camera_angle == 4:
        yaw = 69.59991455078125
        pitch = -33.8000602722168
        cam_dist = 1.8000028133392334
        cam_target_pos = (0.0, 0.0, 0.0)
        roll = 0

    elif camera_angle == 5:
        yaw = 90.800048828125
        pitch = -59.800079345703125
        cam_dist = 1.8000028133392334
        cam_target_pos = (0.0, 0.0, 0.0)
        roll = 0

    elif camera_angle == 6:
        yaw = 90.4000473022461
        pitch = -65.40008544921875
        cam_dist = 2.000002861022949
        cam_target_pos = (0.0, 0.0, 0.0)
        roll = 0

    elif camera_angle == 7:
        yaw = 90.00004577636719
        pitch = -45.4000358581543
        cam_dist = 2.000002861022949
        cam_target_pos = (0.0, 0.0, 0.0)
        roll = 0

    elif camera_angle == 8:
        yaw = 89.60002899169922
        pitch = -17.400007247924805
        cam_dist = 1.4000000953674316
        cam_target_pos = (-0.07712450623512268, 0.05323473736643791, 0.45070940256118774)
        roll = 0

    return p.computeViewMatrixFromYawPitchRoll(cam_target_pos, cam_dist, yaw, pitch, roll, 2)


def compute_projection_matrix():
    fov = 90
    aspect_ratio = VIDEO_WIDTH / VIDEO_HEIGHT
    near_distance = 0.1
    far_distance = 100

    return p.computeProjectionMatrixFOV(fov, aspect_ratio, near_distance, far_distance)


class VideoRecordingManager(ABC, SafeMotionsBase):
    def __init__(self,
                 experiment_name,
                 *vargs,
                 render_video=False,
                 extra_render_modes=None,
                 video_frame_rate=VIDEO_FRAME_RATE,
                 camera_angle=0,
                 **kwargs):
        super().__init__(*vargs, **kwargs)

        # video recording settings
        self._video_recorder = None
        self._render_video = render_video
        time_stamp = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')
        self._video_dir = os.path.join(self._evaluation_dir, self.__class__.__name__, experiment_name, time_stamp)
        self._video_base_path = None
        self._render_modes = RENDER_MODES.copy()
        if extra_render_modes:
            self._render_modes += extra_render_modes
        self._video_frame_rate = video_frame_rate
        self._sim_steps_per_frame = int(1 / (self._video_frame_rate * self._control_time_step))
        self._video_height = VIDEO_HEIGHT
        self._video_width = VIDEO_WIDTH
        self._camera_angle = camera_angle
        self._view_matrix = compute_view_matrix(self._camera_angle)
        self._projection_matrix = compute_projection_matrix()
        self._sim_step_counter = None

    @property
    def metadata(self):
        metadata = {
            'render.modes': self._render_modes,
            'video.frames_per_second': self._video_frame_rate
        }

        return metadata

    def reset(self):
        observation = super().reset()

        self._sim_step_counter = 0

        if self._render_video:
            self._reset_video_recorder()

        return observation

    def close(self):
        if self._video_recorder:
            self._close_video_recorder()

        super().close()

    def _sim_step(self):
        super()._sim_step()
        self._sim_step_counter += 1

        if self._render_video:
            if self._sim_step_counter == self._sim_steps_per_frame:
                self._capture_frame_with_video_recorder()

    def _prepare_for_end_of_episode(self):
        super()._prepare_for_end_of_episode()

        if self._render_video:
            for _ in range(self._video_frame_rate):
                self._capture_frame_with_video_recorder()
            if self._video_recorder:
                self._close_video_recorder()

    def _capture_frame_with_video_recorder(self):
        self._sim_step_counter = 0
        self._video_recorder.capture_frame()

    def _reset_video_recorder(self):
        if self._video_recorder:
            self._close_video_recorder()

        os.makedirs(self._video_dir, exist_ok=True)

        episode_id = self._episode_counter
        self._video_base_path = os.path.join(self._video_dir, "episode_{}".format(episode_id))
        metadata = {'episode_id': episode_id}
        self._video_recorder = VideoRecorder(self, base_path=self._video_base_path, metadata=metadata, enabled=True)

        for _ in range(self._video_frame_rate):
            self._capture_frame_with_video_recorder()

    def _close_video_recorder(self):
        self._video_recorder.close()
        self._video_recorder = []



# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
import os
import time
from abc import abstractmethod
from functools import partial
from pathlib import Path
from threading import Thread

import gym
import numpy as np
import pybullet as p

from safemotions.robot_scene.real_robot_scene import RealRobotScene
from safemotions.robot_scene.simulated_robot_scene import SimRobotScene
from safemotions.utils.trajectory_manager import TrajectoryManager

SIM_TIME_STEP = 1. / 240.
CONTROLLER_TIME_STEP = 1. / 200.
EPISODES_PER_SIMULATION_RESET = 25000  # to avoid out of memory error

# Termination reason
TERMINATION_UNSET = -1
TERMINATION_SUCCESS = 0
TERMINATION_JOINT_LIMITS = 1
TERMINATION_TRAJECTORY_LENGTH = 2


class SafeMotionsBase(gym.Env):
    def __init__(self,
                 evaluation_dir=None,
                 use_real_robot=False,
                 real_robot_debug_mode=False,
                 use_gui=False,
                 control_time_step=None,
                 use_control_rate_sleep=True,
                 use_thread_for_control_rate_sleep=False,
                 pos_limit_factor=1,
                 vel_limit_factor=1,
                 acc_limit_factor=1,
                 jerk_limit_factor=1,
                 torque_limit_factor=1,
                 acceleration_after_max_vel_limit_factor=0.01,
                 eval_new_condition_counter=1,
                 log_obstacle_data=False,
                 save_obstacle_data=False,
                 store_actions=False,
                 online_trajectory_duration=8.0,
                 online_trajectory_time_step=0.1,
                 position_controller_time_constants=None,
                 plot_computed_actual_values=False,
                 plot_actual_torques=False,
                 robot_scene=0,
                 obstacle_scene=0,
                 observed_link_point_scene=0,
                 obstacle_use_computed_actual_values=False,
                 visualize_bounding_spheres=False,
                 use_braking_trajectory_method=False,
                 collision_check_time=0.05,
                 check_braking_trajectory_observed_points=False,
                 check_braking_trajectory_closest_points=True,
                 check_braking_trajectory_torque_limits=False,
                 closest_point_safety_distance=0.1,
                 observed_point_safety_distance=0.1,
                 use_target_points=False,
                 target_point_cartesian_range_scene=0,
                 target_point_radius=0.05,
                 target_point_sequence=0,
                 target_point_reached_reward_bonus=0.0,
                 target_link_name="iiwa_link_7",
                 target_link_offset=None,
                 no_self_collision=False,
                 time_step_fraction_sleep_observation=0.0,
                 seed=None,
                 random_agent=False,
                 **kwargs):

        if target_link_offset is None:
            target_link_offset = [0, 0, 0.0]
        if position_controller_time_constants is None:
            position_controller_time_constants = [0.030, 0.030, 0.030, 0.030, 0.030, 0.030, 0.030]
        if seed is not None:
            np.random.seed(seed)
        if evaluation_dir is None:
            self._evaluation_dir = os.path.join(Path.home(), "safe_motions_evaluation")
        else:
            self._evaluation_dir = os.path.join(evaluation_dir, "safe_motions_evaluation")

        self._use_real_robot = use_real_robot
        self._use_gui = use_gui
        self._use_control_rate_sleep = use_control_rate_sleep
        self._num_physic_clients = 0

        if self._use_gui:
            self._simulation_client_id = p.connect(p.GUI)
            self._num_physic_clients += 1
        else:
            if not self._use_real_robot:
                self._simulation_client_id = p.connect(p.DIRECT)
                self._num_physic_clients += 1
            else:
                self._simulation_client_id = None

        self._obstacle_client_id = p.connect(p.DIRECT)
        self._num_physic_clients += 1

        if control_time_step is None:
            self._control_time_step = CONTROLLER_TIME_STEP if self._use_real_robot else SIM_TIME_STEP
        else:
            self._control_time_step = control_time_step

        self._simulation_time_step = SIM_TIME_STEP
        self._control_step_counter = 0
        self._episode_counter = 0

        self._obstacle_scene = obstacle_scene
        self._observed_link_point_scene = observed_link_point_scene
        self._visualize_bounding_spheres = visualize_bounding_spheres
        self._log_obstacle_data = log_obstacle_data
        self._save_obstacle_data = save_obstacle_data
        self._robot_scene_config = robot_scene
        self._use_braking_trajectory_method = use_braking_trajectory_method
        self._collision_check_time = collision_check_time
        self._check_braking_trajectory_observed_points = check_braking_trajectory_observed_points
        self._check_braking_trajectory_closest_points = check_braking_trajectory_closest_points
        self._check_braking_trajectory_torque_limits = check_braking_trajectory_torque_limits
        self._closest_point_safety_distance = closest_point_safety_distance
        self._observed_point_safety_distance = observed_point_safety_distance
        self._use_target_points = use_target_points
        self._target_point_cartesian_range_scene = target_point_cartesian_range_scene
        self._target_point_radius = target_point_radius
        self._target_point_sequence = target_point_sequence
        self._target_point_reached_reward_bonus = target_point_reached_reward_bonus
        self._no_self_collision = no_self_collision
        self._trajectory_time_step = online_trajectory_time_step
        self._position_controller_time_constants = position_controller_time_constants
        self._plot_computed_actual_values = plot_computed_actual_values
        self._plot_actual_torques = plot_actual_torques
        self._pos_limit_factor = pos_limit_factor
        self._vel_limit_factor = vel_limit_factor
        self._acc_limit_factor = acc_limit_factor
        self._jerk_limit_factor = jerk_limit_factor
        self._torque_limit_factor = torque_limit_factor
        self._acceleration_after_max_vel_limit_factor = acceleration_after_max_vel_limit_factor
        self._online_trajectory_duration = online_trajectory_duration
        self._eval_new_condition_counter = eval_new_condition_counter
        self._store_actions = store_actions
        self._target_link_name = target_link_name
        self._target_link_offset = target_link_offset
        self._real_robot_debug_mode = real_robot_debug_mode
        self._random_agent = random_agent

        self._network_prediction_part_done = None
        self._use_thread_for_control_rate_sleep = use_thread_for_control_rate_sleep
        if self._use_thread_for_control_rate_sleep and not self._use_control_rate_sleep:
            logging.warning("use_thread_for_control_rate_sleep ignored since use_control_rate_sleep == False")
        self._use_movement_thread = (self._use_control_rate_sleep and self._use_thread_for_control_rate_sleep) \
                                    or self._use_real_robot
        self._time_step_fraction_sleep_observation = time_step_fraction_sleep_observation
        # 0..1; fraction of the time step,  the main thread sleeps before getting the next observation;
        # only relevant if self._use_real_robot == True
        if time_step_fraction_sleep_observation != 0:
            logging.info("time_step_fraction_sleep_observation %s", self._time_step_fraction_sleep_observation)
        self._obstacle_use_computed_actual_values = obstacle_use_computed_actual_values
        # use computed actual values to determine the distance between the robot and obstacles and as initial point
        # for torque simulations -> advantage: can be computed in advance, no measurements -> real-time capable
        # disadvantage: controller model might introduce inaccuracies
        if self._use_movement_thread and not self._obstacle_use_computed_actual_values:
            raise ValueError("Real-time execution requires obstacle_use_computed_actual_values to be True")

        if self._use_movement_thread:
            logging.info("Using movement thread")

        self._model_actual_values = self._use_movement_thread or self._obstacle_use_computed_actual_values or \
                                    self._plot_computed_actual_values or (self._use_real_robot and self._use_gui)

        if not self._use_movement_thread and self._control_time_step != self._simulation_time_step:
            raise ValueError("If no movement thread is used, the control time step must equal the control time step "
                             "of the obstacle client")

        self._start_position = None
        self._start_velocity = None
        self._start_acceleration = None
        self._position_deviation = None
        self._acceleration_deviation = None
        self._current_trajectory_point_index = None
        self._trajectory_successful = None
        self._total_reward = None
        self._episode_length = None
        self._action_list = []
        self._last_action = None
        self._termination_reason = TERMINATION_UNSET
        self._movement_thread = None
        self._braking = False

        self._init_simulation()

    def _init_simulation(self):
        # reset the physics engine
        for i in range(self._num_physic_clients):
            p.resetSimulation(physicsClientId=i)  # to free memory

        # robot scene settings
        robot_scene_parameters = {'simulation_client_id': self._simulation_client_id,
                                  'simulation_time_step': self._simulation_time_step,
                                  'obstacle_client_id': self._obstacle_client_id,
                                  'trajectory_time_step': self._trajectory_time_step,
                                  'use_real_robot': self._use_real_robot,
                                  'robot_scene': self._robot_scene_config,
                                  'obstacle_scene': self._obstacle_scene,
                                  'observed_link_point_scene': self._observed_link_point_scene,
                                  'log_obstacle_data': self._log_obstacle_data,
                                  'visualize_bounding_spheres': self._visualize_bounding_spheres,
                                  'use_braking_trajectory_method': self._use_braking_trajectory_method,
                                  'collision_check_time': self._collision_check_time,
                                  'check_braking_trajectory_observed_points':
                                      self._check_braking_trajectory_observed_points,
                                  'check_braking_trajectory_closest_points':
                                      self._check_braking_trajectory_closest_points,
                                  'check_braking_trajectory_torque_limits':
                                      self._check_braking_trajectory_torque_limits,
                                  'closest_point_safety_distance': self._closest_point_safety_distance,
                                  'observed_point_safety_distance': self._observed_point_safety_distance,
                                  'use_target_points': self._use_target_points,
                                  'target_point_cartesian_range_scene': self._target_point_cartesian_range_scene,
                                  'target_point_radius': self._target_point_radius,
                                  'target_point_sequence': self._target_point_sequence,
                                  'target_point_reached_reward_bonus': self._target_point_reached_reward_bonus,
                                  'no_self_collision': self._no_self_collision,
                                  'target_link_name': self._target_link_name,
                                  'target_link_offset': self._target_link_offset,
                                  'pos_limit_factor': self._pos_limit_factor,
                                  'vel_limit_factor': self._vel_limit_factor,
                                  'acc_limit_factor': self._acc_limit_factor,
                                  'jerk_limit_factor': self._jerk_limit_factor,
                                  'torque_limit_factor': self._torque_limit_factor
                                  }

        if self._use_real_robot:
            self._robot_scene = RealRobotScene(real_robot_debug_mode=self._real_robot_debug_mode,
                                               **robot_scene_parameters)
        else:
            self._robot_scene = SimRobotScene(**robot_scene_parameters)

        self._num_manip_joints = self._robot_scene.num_manip_joints
        # trajectory manager settings
        self._trajectory_manager = TrajectoryManager(trajectory_time_step=self._trajectory_time_step,
                                                     trajectory_duration=self._online_trajectory_duration,
                                                     obstacle_wrapper=self._robot_scene.obstacle_wrapper)

        self._robot_scene.compute_actual_joint_limits()
        self._control_steps_per_action = int(round(self._trajectory_time_step / self._control_time_step))
        self._obstacle_client_update_steps_per_action = int(round(self._trajectory_time_step /
                                                                  self._simulation_time_step))

        logging.info("Trajectory time step: " + str(self._trajectory_time_step))

        # calculate model coefficients to estimate actual values if required
        if self._model_actual_values:
            self._trajectory_manager.compute_controller_model_coefficients(self._position_controller_time_constants,
                                                                           self._simulation_time_step)

        self._zero_joint_vector = [0.0] * self._num_manip_joints

        if self._use_movement_thread or (self._use_gui and self._use_control_rate_sleep):
            try:
                import rospy
                rospy.init_node("safe_motions_control_rate", anonymous=True, disable_signals=True)
                self._control_rate = rospy.Rate(1. / self._control_time_step)
            except ImportError as _:
                logging.warning("Could not find rospy / a ROS installation. Using time.sleep instead of rospy.Rate. "
                                "Timing will be less accurate.")
                control_rate_sleep = partial(time.sleep, self._control_time_step)
                self._control_rate = type("control_rate", (object, ), {"sleep": lambda: control_rate_sleep()})
        else:
            self._control_rate = None

        for i in range(self._num_physic_clients):
            p.setGravity(0, 0, -9.81, physicsClientId=i)
            p.setPhysicsEngineParameter(numSolverIterations=150, physicsClientId=i)
            p.setTimeStep(self._simulation_time_step, physicsClientId=i)

    def reset(self):

        self._episode_counter += 1

        if self._episode_counter % EPISODES_PER_SIMULATION_RESET == 0:
            self._init_simulation()

        self._control_step_counter = 0

        self._total_reward = 0
        self._episode_length = 0
        self._trajectory_successful = True
        self._current_trajectory_point_index = 0
        self._action_list = []

        self._network_prediction_part_done = False

        get_new_setup = (((self._episode_counter-1) % self._eval_new_condition_counter) == 0)

        self._robot_scene.obstacle_wrapper.reset_obstacles()
        self._trajectory_manager.reset(get_new_trajectory=get_new_setup)
        self._start_position = np.array(self._get_trajectory_start_position())
        self._start_velocity = np.array(self._zero_joint_vector)
        self._start_acceleration = np.array(self._zero_joint_vector)
        if self._use_real_robot:
            logging.info("Starting position: %s", self._start_position)
        else:
            logging.debug("Starting position: %s", self._start_position)
        self._robot_scene.pose_manipulator(self._start_position)
        self._robot_scene.obstacle_wrapper.reset(start_position=self._start_position)
        self._robot_scene.obstacle_wrapper.update(target_position=self._start_position,
                                                  target_velocity=self._start_velocity,
                                                  target_acceleration=self._start_acceleration,
                                                  actual_position=self._start_position,
                                                  actual_velocity=self._start_velocity)

        self._reset_plotter(self._start_position)
        self._add_computed_actual_position_to_plot(self._start_position, self._zero_joint_vector,
                                                   self._zero_joint_vector)

        if self._plot_actual_torques and not self._use_real_robot:
            # the initial torques are not zero due to gravity
            self._robot_scene.set_motor_control(target_positions=self._start_position,
                                                physics_client_id=self._simulation_client_id)
            p.stepSimulation(physicsClientId=self._simulation_client_id)
            actual_joint_torques = self._robot_scene.get_actual_joint_torques()
            self._add_actual_torques_to_plot(actual_joint_torques)
        else:
            self._add_actual_torques_to_plot(self._zero_joint_vector)

        self._calculate_norm_acc_range(self._start_position, self._start_velocity, self._start_acceleration,
                                       self._current_trajectory_point_index)

        self._termination_reason = TERMINATION_UNSET
        self._last_action = None
        self._network_prediction_part_done = False
        self._movement_thread = None
        self._braking = False

        if self._control_rate is not None:
            # reset control rate timer
            self._control_rate.sleep()

        return None

    def step(self, action):
        self._episode_length += 1

        if self._random_agent:
            action = np.random.uniform(-1, 1, self.action_space.shape)
            # overwrite the desired action with a random action

        if self._store_actions:
            self._action_list.append(action)

        controller_setpoints, obstacle_client_update_setpoints, action_info, robot_stopped = \
            self._compute_controller_setpoints_from_action(action)

        for i in range(len(controller_setpoints['positions'])):
            self._add_generated_trajectory_control_point(controller_setpoints['positions'][i],
                                                         controller_setpoints['velocities'][i],
                                                         controller_setpoints['accelerations'][i])

        for i in range(len(obstacle_client_update_setpoints['positions'])):

            if self._model_actual_values:

                last_position_setpoint = self._start_position if i == 0 else obstacle_client_update_setpoints[
                    'positions'][i - 1]
                computed_position_is = self._trajectory_manager.model_position_controller_to_compute_actual_position(
                    current_position_setpoint=obstacle_client_update_setpoints['positions'][i],
                    last_position_setpoint=last_position_setpoint)
                computed_velocity_is = (np.array(computed_position_is) - np.array(
                    self._get_computed_actual_trajectory_control_point(-1))) / self._simulation_time_step
                computed_acceleration_is = (computed_velocity_is - np.array(
                    self._get_computed_actual_trajectory_control_point(-1,
                                                                       key='velocities'))) / self._simulation_time_step

                self._add_computed_actual_trajectory_control_point(list(computed_position_is),
                                                                   list(computed_velocity_is),
                                                                   list(computed_acceleration_is))
                self._add_computed_actual_position_to_plot(computed_position_is, computed_velocity_is,
                                                           computed_acceleration_is)

                if self._robot_scene.obstacle_wrapper is not None:
                    if self._use_movement_thread or self._obstacle_use_computed_actual_values:
                        self._robot_scene.obstacle_wrapper.update(
                            target_position=obstacle_client_update_setpoints['positions'][i],
                            target_velocity=obstacle_client_update_setpoints['velocities'][i],
                            target_acceleration=obstacle_client_update_setpoints['accelerations'][
                                                                       i],
                            actual_position=computed_position_is,
                            actual_velocity=computed_velocity_is)

        if self._use_movement_thread:
            movement_thread = Thread(target=self._execute_robot_movement,
                                     kwargs=dict(controller_setpoints=controller_setpoints))
            if self._movement_thread is not None:
                self._movement_thread.join()
            movement_thread.start()
            self._movement_thread = movement_thread
            movement_info = {}
        else:
            self._movement_thread = None
            movement_info = self._execute_robot_movement(controller_setpoints=controller_setpoints)

        self._start_position = np.array(obstacle_client_update_setpoints['positions'][-1])
        self._start_velocity = np.array(obstacle_client_update_setpoints['velocities'][-1])
        self._start_acceleration = np.array(obstacle_client_update_setpoints['accelerations'][-1])
        self._add_generated_trajectory_point(self._start_position, self._start_velocity, self._start_acceleration)

        self._current_trajectory_point_index += 1
        self._last_action = action  # store the last action for reward calculation

        self._calculate_norm_acc_range(self._start_position, self._start_velocity, self._start_acceleration,
                                       self._current_trajectory_point_index)

        # sleep for a specified part of the time_step before getting the observation
        if self._time_step_fraction_sleep_observation != 0:
            time.sleep(self._trajectory_time_step * self._time_step_fraction_sleep_observation)

        observation, reward, done, info = self._process_action_outcome(movement_info, action_info)

        if not self._network_prediction_part_done:
            self._total_reward += reward
        else:
            done = True

        if done:
            self._network_prediction_part_done = True

        if not self._network_prediction_part_done:
            self._prepare_for_next_action()
        else:
            if not self._use_real_robot or robot_stopped:
                if self._trajectory_successful:
                    self._termination_reason = TERMINATION_SUCCESS

                else:
                    if self._episode_length == self._trajectory_manager.trajectory_length - 1 and \
                            self._termination_reason == TERMINATION_UNSET:
                        self._termination_reason = TERMINATION_TRAJECTORY_LENGTH

                if self._movement_thread is not None:
                    self._movement_thread.join()
                self._robot_scene.prepare_for_end_of_episode()
                self._prepare_for_end_of_episode()
                observation, reward, _, info = self._process_end_of_episode(observation, reward, done, info)

                if self._store_actions:
                    self._store_action_list()

            else:
                self._braking = True  # slow down the robot prior to stopping the episode
                done = False

        return observation, reward, done, dict(info)

    def _execute_robot_movement(self, controller_setpoints):
        # executed in real-time if required
        actual_joint_torques_rel_abs_list = []
        for i in range(len(controller_setpoints['positions'])):
            if not self._use_real_robot:
                curr_position_is = self._robot_scene.get_actual_joint_positions()
                curr_velocity_is = (np.array(curr_position_is) - np.array(
                    self._get_measured_actual_trajectory_control_point(-1))) / self._control_time_step
                curr_acceleration_is = (curr_velocity_is - np.array(
                    self._get_measured_actual_trajectory_control_point(-1, key='velocities'))) / self._control_time_step
                self._add_measured_actual_trajectory_control_point(list(curr_position_is), list(curr_velocity_is),
                                                                   list(curr_acceleration_is))
                self._add_actual_position_to_plot(curr_position_is)

            if self._control_rate is not None:
                self._control_rate.sleep()

            self._robot_scene.set_motor_control(controller_setpoints['positions'][i],
                                                computed_position_is=controller_setpoints['positions'][i],
                                                computed_velocity_is=controller_setpoints['velocities'][i])

            if not self._use_real_robot:
                self._sim_step()
                actual_joint_torques = self._robot_scene.get_actual_joint_torques()
                actual_joint_torques_rel_abs = np.abs(normalize_joint_values(actual_joint_torques,
                                                                             self._robot_scene.max_torques))
                actual_joint_torques_rel_abs_list.append(actual_joint_torques_rel_abs)

                if self._plot_actual_torques:
                    self._add_actual_torques_to_plot(actual_joint_torques)
            else:
                actual_joint_torques_rel_abs_list.append(self._zero_joint_vector)

            if self._robot_scene.obstacle_wrapper is not None:
                if not self._use_movement_thread and not self._obstacle_use_computed_actual_values:
                    actual_position, actual_velocity = self._robot_scene.get_actual_joint_position_and_velocity()

                    self._robot_scene.obstacle_wrapper.update(target_position=controller_setpoints['positions'][i],
                                                              target_velocity=controller_setpoints['velocities'][i],
                                                              target_acceleration=controller_setpoints['accelerations'][
                                                                  i],
                                                              actual_position=actual_position,
                                                              actual_velocity=actual_velocity)

        movement_info = {'average': {}, 'max': {}}

        # add torque info to movement_info
        torque_violation = 0.0
        actual_joint_torques_rel_abs_swap = np.swapaxes(actual_joint_torques_rel_abs_list, 0, 1)
        for j in range(self._num_manip_joints):
            movement_info['average']['joint_{}_torque_abs'.format(j)] = np.mean(actual_joint_torques_rel_abs_swap[j])
            actual_joint_torques_rel_abs_max = np.max(actual_joint_torques_rel_abs_swap[j])
            movement_info['max']['joint_{}_torque_abs'.format(j)] = actual_joint_torques_rel_abs_max
            if actual_joint_torques_rel_abs_max > 1.001:
                torque_violation = 1.0
                logging.warning("Torque Violation: t = " + str(
                    (self._episode_length - 1) * self._trajectory_time_step) + " Joint: " + str(
                    j) + " Rel Torque:" + str(actual_joint_torques_rel_abs_max))

        movement_info['max']['joint_torque_violation'] = torque_violation
        movement_info['average']['joint_torque_violation'] = torque_violation

        return movement_info

    def close(self):
        self._robot_scene.disconnect()

    @abstractmethod
    def render(self, mode="human"):
        raise NotImplementedError()

    @abstractmethod
    def _get_observation(self):
        raise NotImplementedError()

    @abstractmethod
    def _get_reward(self):
        raise NotImplementedError()

    @abstractmethod
    def _compute_controller_setpoints_from_action(self, action):
        raise NotImplementedError()

    @abstractmethod
    def _interpolate_position(self, step):
        raise NotImplementedError()

    @abstractmethod
    def _interpolate_velocity(self, step):
        raise NotImplementedError()

    @abstractmethod
    def _interpolate_acceleration(self, step):
        raise NotImplementedError()

    @abstractmethod
    def _process_action_outcome(self, base_info, action_info):
        raise NotImplementedError()

    @abstractmethod
    def _process_end_of_episode(self, observation, reward, done, info):
        raise NotImplementedError()

    def _sim_step(self):
        p.stepSimulation(physicsClientId=self._simulation_client_id)
        self._control_step_counter += 1

    def _prepare_for_next_action(self):
        return

    def _prepare_for_end_of_episode(self):
        return

    @abstractmethod
    def _store_action_list(self):
        raise NotImplementedError()

    def _check_termination(self):
        done = False
        if self._trajectory_manager.is_trajectory_finished(self._current_trajectory_point_index):
            done = True

        return done

    @property
    def norm_acc_range(self):
        return self._get_norm_acc_range()

    @property
    def trajectory_time_step(self):
        return self._trajectory_time_step

    @abstractmethod
    def _get_norm_acc_range(self):
        pass

    @abstractmethod
    def _reset_plotter(self, initial_joint_position):
        pass

    @abstractmethod
    def _add_actual_position_to_plot(self, actual_joint_position):
        pass

    @abstractmethod
    def _add_computed_actual_position_to_plot(self, computed_position_is, computed_velocity_is,
                                              computed_acceleration_is):
        pass

    @abstractmethod
    def _add_actual_torques_to_plot(self, actual_torques):
        pass

    @abstractmethod
    def _calculate_norm_acc_range(self, start_position, start_velocity, start_acceleration, trajectory_point_index):
        pass

    def _get_trajectory_start_position(self):
        return self._trajectory_manager.get_trajectory_start_position()

    def _get_generated_trajectory_point(self, index, key='positions'):
        return self._trajectory_manager.get_generated_trajectory_point(index, key)

    def _get_measured_actual_trajectory_control_point(self, index, key='positions'):
        return self._trajectory_manager.get_measured_actual_trajectory_control_point(index, key)

    def _get_computed_actual_trajectory_control_point(self, index, key='positions'):
        return self._trajectory_manager.get_computed_actual_trajectory_control_point(index, key)

    def _get_generated_trajectory_control_point(self, index, key='positions'):
        return self._trajectory_manager.get_generated_trajectory_control_point(index, key)

    def _add_generated_trajectory_point(self, position, velocity, acceleration):
        self._trajectory_manager.add_generated_trajectory_point(position, velocity, acceleration)

    def _add_measured_actual_trajectory_control_point(self, position, velocity, acceleration):
        self._trajectory_manager.add_measured_actual_trajectory_control_point(position, velocity, acceleration)

    def _add_computed_actual_trajectory_control_point(self, position, velocity, acceleration):
        self._trajectory_manager.add_computed_actual_trajectory_control_point(position, velocity, acceleration)

    def _add_generated_trajectory_control_point(self, position, velocity, acceleration):
        self._trajectory_manager.add_generated_trajectory_control_point(position, velocity, acceleration)


def normalize_joint_values(values, joint_limits):
    return list(np.array(values) / np.array(joint_limits))



# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
import numpy as np
import os
import inspect
import errno
import datetime
import pybullet as p
from collections import defaultdict
from glob import glob
from itertools import chain
from safemotions.envs.decorators import actions, observations, rewards, video_recording

project_dir = os.path.dirname(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))))


class SafeMotionsEnv(actions.AccelerationPredictionBoundedJerkAccVelPos,
                     observations.SafeObservation,
                     rewards.TargetPointReachingReward,
                     video_recording.VideoRecordingManager):
    def __init__(self,
                 experiment_name,
                 *vargs,
                 **kwargs):
        super().__init__(experiment_name, *vargs, **kwargs)
        self._experiment_name = experiment_name
        self._timestamp = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')

    def render(self, mode="human"):
        if mode == "human":
            return np.array([])
        else:
            (_, _, image, _, _) = p.getCameraImage(width=self._video_width, height=self._video_height,
                                                   renderer=p.ER_BULLET_HARDWARE_OPENGL,
                                                   viewMatrix=self._view_matrix,
                                                   projectionMatrix=self._projection_matrix)
            image = np.reshape(image, (self._video_height, self._video_width, 4))
            image = np.uint8(image[:, :, :3])

            return np.array(image)

    def _process_action_outcome(self, base_info, action_info):
        reward, reward_info = self._get_reward()
        observation, observation_info = self._get_observation()
        done = self._check_termination()
        info = defaultdict(dict)

        for k, v in chain(base_info.items(), action_info.items(), observation_info.items(), reward_info.items()):
            info[k] = {**info[k], **v}

        return observation, reward, done, info

    def _process_end_of_episode(self, observation, reward, done, info):
        info.update(trajectory_length=self._trajectory_manager.trajectory_length)
        info.update(episode_length=self._episode_length)
        info['termination_reason'] = self._termination_reason

        # get info from obstacle wrapper
        obstacle_info = self._robot_scene.obstacle_wrapper.get_info_and_print_stats()
        info = dict(info, **obstacle_info)  # concatenate dicts
        self._display_plot()
        self._save_plot(self.__class__.__name__, self._experiment_name)

        return observation, reward, done, info

    def _store_action_list(self):
        action_dict = {'actions': np.asarray(self._action_list).tolist()}
        eval_dir = os.path.join(project_dir, "evalution", "action_logs",
                                self.__class__.__name__, self._experiment_name, self._timestamp)

        if not os.path.exists(eval_dir):
            try:
                os.makedirs(eval_dir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

        with open(os.path.join(eval_dir, "episode_" + str(self._episode_counter) + ".json"), 'w') as f:
            f.write(json.dumps(action_dict))
            f.flush()

    def _create_metadata(self):
        return {'episode_id': self._episode_counter, 'trajectory_length': self._trajectory_manager.trajectory_length,
                'episode_length': self._episode_length, 'total_reward': self._total_reward}

    def _close_video_recorder(self):
        super()._close_video_recorder()
        self._adapt_rendering_metadata()
        self._rename_output_files()

    def _adapt_rendering_metadata(self):
        metadata_ext = ".meta.json"
        metadata_file = self._video_base_path + metadata_ext

        with open(metadata_file, 'r') as f:
            metadata_json = json.load(f)

            encoder_metadata = metadata_json.pop('encoder_version', None)
            if encoder_metadata:
                metadata_json.update(encoder=encoder_metadata['backend'])

            metadata_json.update(trajectory_length=self._trajectory_manager.trajectory_length)
            metadata_json.update(episode_length=self._episode_length)
            metadata_json.update(total_reward=round(self._total_reward, 3))

        with open(metadata_file, 'w') as f:
            f.write(json.dumps(metadata_json, indent=4))
            f.close()

    def _rename_output_files(self):
        output_file = glob(self._video_base_path + ".*")

        for file in output_file:
            dir_path, file_name = os.path.split(file)
            name, extension = os.path.splitext(file_name)
            new_name = "_".join(map(str, [name, self._episode_length,
                                          self._robot_scene.obstacle_wrapper.get_num_target_points_reached(),
                                          round(self._total_reward, 3)]))
            new_file_name = new_name + extension
            os.rename(file, os.path.join(dir_path, new_file_name))



# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np


def clip_index(index, list_length):
    if index < 0 and abs(index) > list_length:
        return 0
    if index > 0 and index > list_length - 1:
        return -1
    else:
        return index


class TrajectoryManager(object):

    def __init__(self,
                 trajectory_time_step,
                 trajectory_duration,
                 obstacle_wrapper,
                 **kwargs):

        self._trajectory_time_step = trajectory_time_step
        self._trajectory_duration = trajectory_duration
        self._obstacle_wrapper = obstacle_wrapper

        self._trajectory_start_position = None
        self._trajectory_length = None
        self._num_manip_joints = None
        self._zero_joint_vector = None
        self._generated_trajectory = None
        self._measured_actual_trajectory_control_points = None
        self._computed_actual_trajectory_control_points = None
        self._generated_trajectory_control_points = None

        self._controller_model_coefficient_a = None
        self._controller_model_coefficient_b = None

    @property
    def trajectory_time_step(self):
        return self._trajectory_time_step

    @property
    def trajectory_length(self):
        return self._trajectory_length

    def reset(self, get_new_trajectory=True):
        if get_new_trajectory:
            self._trajectory_start_position = self._get_new_trajectory_start_position()
        self._trajectory_length = int(self._trajectory_duration / self._trajectory_time_step) + 1
        self._num_manip_joints = len(self._trajectory_start_position)
        self._zero_joint_vector = [0.0] * self._num_manip_joints
        self._generated_trajectory = {'positions': [self.get_trajectory_start_position()],
                                      'velocities': [self._zero_joint_vector],
                                      'accelerations': [self._zero_joint_vector]}
        self._measured_actual_trajectory_control_points = {'positions': [self.get_trajectory_start_position()],
                                                           'velocities': [self._zero_joint_vector],
                                                           'accelerations': [self._zero_joint_vector]}
        self._computed_actual_trajectory_control_points = {'positions': [self.get_trajectory_start_position()],
                                                           'velocities': [self._zero_joint_vector],
                                                           'accelerations': [self._zero_joint_vector]}
        self._generated_trajectory_control_points = {'positions': [self.get_trajectory_start_position()],
                                                     'velocities': [self._zero_joint_vector],
                                                     'accelerations': [self._zero_joint_vector]}

    def get_trajectory_start_position(self):
        return self._trajectory_start_position

    def _denormalize(self, norm_value, value_range):
        actual_value = value_range[0] + 0.5 * (norm_value + 1) * (value_range[1] - value_range[0])
        return actual_value

    def get_generated_trajectory_point(self, index, key='positions'):
        i = clip_index(index, len(self._generated_trajectory[key]))

        return self._generated_trajectory[key][i]

    def get_measured_actual_trajectory_control_point(self, index, key='positions'):
        i = clip_index(index, len(self._measured_actual_trajectory_control_points[key]))

        return self._measured_actual_trajectory_control_points[key][i]

    def get_computed_actual_trajectory_control_point(self, index, key='positions'):
        i = clip_index(index, len(self._computed_actual_trajectory_control_points[key]))

        return self._computed_actual_trajectory_control_points[key][i]

    def get_generated_trajectory_control_point(self, index, key='positions'):
        i = clip_index(index, len(self._generated_trajectory_control_points[key]))

        return self._generated_trajectory_control_points[key][i]

    def add_generated_trajectory_point(self, positions, velocities, accelerations):
        self._generated_trajectory['positions'].append(positions)
        self._generated_trajectory['velocities'].append(velocities)
        self._generated_trajectory['accelerations'].append(accelerations)

    def add_measured_actual_trajectory_control_point(self, positions, velocities, accelerations):
        self._measured_actual_trajectory_control_points['positions'].append(positions)
        self._measured_actual_trajectory_control_points['velocities'].append(velocities)
        self._measured_actual_trajectory_control_points['accelerations'].append(accelerations)

    def add_computed_actual_trajectory_control_point(self, positions, velocities, accelerations):
        self._computed_actual_trajectory_control_points['positions'].append(positions)
        self._computed_actual_trajectory_control_points['velocities'].append(velocities)
        self._computed_actual_trajectory_control_points['accelerations'].append(accelerations)

    def add_generated_trajectory_control_point(self, positions, velocities, accelerations):
        self._generated_trajectory_control_points['positions'].append(positions)
        self._generated_trajectory_control_points['velocities'].append(velocities)
        self._generated_trajectory_control_points['accelerations'].append(accelerations)

    def compute_controller_model_coefficients(self, time_constants, sampling_time):
        self._controller_model_coefficient_a = 1 + (2 * np.array(time_constants) / sampling_time)
        self._controller_model_coefficient_b = 1 - (2 * np.array(time_constants) / sampling_time)

    def model_position_controller_to_compute_actual_position(self, current_position_setpoint, last_position_setpoint):
        # models the position controller as a discrete transfer function and returns the
        # computed actual position, given the next position setpoint and previous computed actual positions
        # the controller is modelled as a first order low-pass with a (continuous) transfer function of
        #  G(s) = 1 / (1 + T * s)
        # the transfer function is discretized using Tustins approximation: s = 2 / Ta * (z - 1) / (z + 1)
        # the following difference equation can be derived:
        # y_n = 1/a * (x_n + x_n_minus_one - b * y_n_minus_one) with a = 1 + (2 * T / Ta) and b = 1 - (2 * T / Ta)

        x_n = np.array(current_position_setpoint)
        x_n_minus_one = np.array(last_position_setpoint)
        y_n_minus_one = np.array(self.get_computed_actual_trajectory_control_point(-1))
        computed_actual_position = 1 / self._controller_model_coefficient_a * \
                                   (x_n + x_n_minus_one - self._controller_model_coefficient_b * y_n_minus_one)
        return computed_actual_position

    def is_trajectory_finished(self, index):
        return index >= self._trajectory_length - 1

    def _get_new_trajectory_start_position(self):
        return self._obstacle_wrapper.get_starting_point_joint_pos()






import os
from setuptools import find_packages
from setuptools import setup

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    readme_file = f.read()

setup(name='safemotions',
      version='0.2.4',
      packages=find_packages(),
      include_package_data=True,
      author='Jonas C. Kiemel',
      author_email='jonas.kiemel@kit.edu',
      url='https://github.com/translearn/safemotions',
      description='Learning Collision-free and Torque-limited Robot Trajectories based on Alternative Safe Behaviors.',
      long_description=readme_file,
      long_description_content_type='text/markdown',
      python_requires='>=3.5, <3.9',
      install_requires=[
          'numpy',
          'klimits',
          'matplotlib',
          'pybullet',
          'gym',
          'ray[tune]==0.8.4',
          'ray[rllib]==0.8.4',
          'tensorflow'
      ],
      )



# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np
import pybullet as p
from safemotions.robot_scene.robot_scene_base import RobotSceneBase


class SimRobotScene(RobotSceneBase):
    def __init__(self,
                 **kwargs):

        super().__init__(**kwargs)

    def pose_manipulator(self, joint_positions):
        for i in range(self._num_manip_joints):
            p.resetJointState(self._robot_id,
                              jointIndex=self._manip_joint_indices[i],
                              targetValue=joint_positions[i],
                              targetVelocity=0,
                              physicsClientId=self._simulation_client_id)

    def get_actual_joint_position_and_velocity(self, manip_joint_indices=None):
        if manip_joint_indices is None:
            manip_joint_indices = self._manip_joint_indices
        # return the actual joint position and velocity for the specified joint indices from the physicsClient
        joint_states = p.getJointStates(self._robot_id, manip_joint_indices,
                                        physicsClientId=self._simulation_client_id)

        joint_states_swap = np.swapaxes(np.array(joint_states, dtype=object), 0, 1)

        return joint_states_swap[0], joint_states_swap[1]



#!/usr/bin/env python

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import argparse
import json
import os
import sys
import inspect
import ray
import datetime
import time
import logging
import numpy as np
import errno
from ray.rllib import rollout
from ray import tune
from pathlib import Path
current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.append(os.path.dirname(current_dir))
from safemotions.envs.safe_motions_env import SafeMotionsEnv

# Termination reason
TERMINATION_UNSET = -1
TERMINATION_SUCCESS = 0
TERMINATION_JOINT_LIMITS = 1
TERMINATION_TRAJECTORY_LENGTH = 2


def np_encoder(object):
    if isinstance(object, np.generic):
        return object.item()


def rollout_manually(agent, evaluation_dir):
    env = agent.workers.local_worker().env
    episodes_sampled = 0
    timestamp = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')
    experiment_name = os.path.join(env_config['experiment_name'], timestamp)

    if args.store_metrics:
        evaluation_dir = os.path.join(evaluation_dir, "trajectory_logs")
        if args.use_real_robot:
            evaluation_dir = os.path.join(evaluation_dir, "real")
        else:
            evaluation_dir = os.path.join(evaluation_dir, "sim")
        if not os.path.exists(os.path.join(evaluation_dir, experiment_name)):
            try:
                os.makedirs(os.path.join(evaluation_dir, experiment_name))
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise
        with open(os.path.join(evaluation_dir, experiment_name, 'config.json'), 'w') as f:
            f.write(json.dumps(vars(args)))
            f.flush()

    episode_computation_time_list = []
    start = time.time()

    while True:
        if args.episodes:
            if episodes_sampled >= args.episodes:
                break

        if args.use_real_robot:
            metric_file = str(episodes_sampled) + '_' + timestamp + "_real" + ".json"
        else:
            metric_file = str(episodes_sampled) + '_' + timestamp + "_sim" + ".json"

        obs = env.reset()
        done = False
        reward_total = 0.0
        episode_info = {}
        steps = -1
        start_episode_timer = time.time()
        while not done:
            steps = steps + 1
            action = agent.compute_action(obs)

            if args.store_metrics:
                next_obs, reward, done, info = env.step(action)
                if episode_info:
                    for key, value in episode_info.items():
                        for key_2, value_2 in episode_info[key].items():
                            episode_info[key][key_2].append(info[key][key_2])
                else:
                    for key, value in info.items():
                        episode_info[key] = {}
                        if isinstance(info[key], dict):
                            for key_2, value_2 in info[key].items():
                                episode_info[key][key_2] = [info[key][key_2]]
            else:
                next_obs, reward, done, _ = env.step(action)

            reward_total += reward
            logging.debug("Observation %s: %s", steps, obs)
            logging.debug("Action %s: %s", steps, action)
            logging.debug("Reward %s: %s", steps, reward)
            obs = next_obs

        end_episode_timer = time.time()
        episode_computation_time = end_episode_timer - start_episode_timer
        logging.info("Last episode took %s seconds", episode_computation_time)
        episode_computation_time_list.append(episode_computation_time)
        logging.info("Trajectory duration: %s seconds", (steps + 1) * env.trajectory_time_step)

        if args.store_metrics:
            episode_info['reward'] = float(reward_total)
            episode_info['episode_length'] = int(info['episode_length'])
            episode_info['trajectory_length'] = int(info['trajectory_length'])
            for key, value in info.items():
                if key.startswith("obstacles"):
                    episode_info[key] = value

            for key, value in episode_info['max'].items():
                episode_info['max'][key] = float(np.max(np.array(value)))
            for key, value in episode_info['average'].items():
                episode_info['average'][key] = float(np.mean(np.array(value)))
            for key, value in episode_info['min'].items():
                episode_info['min'][key] = float(np.min(np.array(value)))

            if info['termination_reason'] == TERMINATION_SUCCESS:
                episode_info['success_rate'] = int(1.0)
            else:
                episode_info['success_rate'] = int(0.0)

            if info['termination_reason'] == TERMINATION_JOINT_LIMITS:
                episode_info['joint_limit_violation_termination_rate'] = int(1.0)
            else:
                episode_info['joint_limit_violation_termination_rate'] = int(0.0)

            with open(os.path.join(evaluation_dir, experiment_name, metric_file), 'w') as f:
                f.write(json.dumps(episode_info, default=np_encoder))
                f.flush()

        episodes_sampled += 1
        logging.info("Episode reward: %s", reward_total)

    env.close()
    end = time.time()
    logging.info("Computed %s episodes in %s seconds.", len(episode_computation_time_list), end - start)
    logging.info("Mean computation time: %s seconds, Max computation time: %s seconds.",
                 np.mean(episode_computation_time_list),
                 np.max(episode_computation_time_list))


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument('--name', type=str, default=None,
                        help="The name of the evaluation.")
    parser.add_argument('--evaluation_dir', type=str, default=None)
    parser.add_argument('--checkpoint', type=str, required=True,
                        help="Path to the checkpoint for evaluation.")
    parser.add_argument('--episodes', type=int, default=20,
                        help="The number of episodes for evaluation.")
    parser.add_argument('--render', action='store_true', default=False,
                        help="Whether or not to render videos of the rollouts.")
    parser.add_argument('--camera_angle', type=int, default=0)
    parser.add_argument('--use_real_robot', action='store_true', default=None)
    parser.add_argument('--real_robot_debug_mode', dest='real_robot_debug_mode', action='store_true', default=False)
    parser.add_argument('--use_gui', action='store_true', default=False)
    parser.add_argument('--online_trajectory_duration', type=float, default=None)
    parser.add_argument('--store_metrics', action='store_true', default=False)
    parser.add_argument('--plot_trajectory', action='store_true', default=False)
    parser.add_argument('--save_trajectory_plot', action='store_true', default=False)
    parser.add_argument('--plot_acc_limits', action='store_true', default=False)
    parser.add_argument('--plot_actual_values', action='store_true', default=False)
    parser.add_argument('--plot_actual_torques', action='store_true', default=False)
    parser.add_argument('--plot_computed_actual_values', action='store_true', default=False)
    parser.add_argument('--plot_joint', type=json.loads, default=None)
    parser.add_argument('--store_actions', action='store_true', default=False)
    parser.add_argument('--log_obstacle_data', action='store_true', default=False)
    parser.add_argument('--obstacle_scene', type=int, default=None)
    parser.add_argument('--seed', type=int, default=None)
    parser.add_argument('--random_agent', action='store_true', default=False)
    parser.add_argument('--no_self_collision', action='store_true', default=False)
    parser.add_argument('--use_thread_for_control_rate_sleep', action='store_true', default=False)
    parser.add_argument('--control_time_step', type=float, default=None)
    parser.add_argument('--time_step_fraction_sleep_observation', type=float, default=None)

    args = parser.parse_args()

    if args.evaluation_dir is None:
        evaluation_dir = os.path.join(Path.home(), "safe_motions_evaluation")
    else:
        evaluation_dir = os.path.join(args.evaluation_dir, "safe_motions_evaluation")

    if not os.path.isdir(args.checkpoint) and not os.path.isfile(args.checkpoint):
        checkpoint_path = os.path.join(current_dir, "trained_networks", args.checkpoint)
    else:
        checkpoint_path = args.checkpoint

    if os.path.isdir(checkpoint_path):
        if os.path.basename(checkpoint_path) == "checkpoint":
            checkpoint_path = os.path.join(checkpoint_path, "checkpoint")
        else:
            checkpoint_path = os.path.join(checkpoint_path, "checkpoint", "checkpoint")

    if not os.path.isfile(checkpoint_path):
        raise ValueError("Could not find checkpoint {}".format(checkpoint_path))

    params_dir = os.path.dirname(os.path.dirname(checkpoint_path))
    params_path = os.path.join(params_dir, "params.json")

    with open(params_path) as params_file:
        checkpoint_config = json.load(params_file)
        env_config = checkpoint_config['env_config']

    if args.name is not None:
        env_config['experiment_name'] = args.name

    if args.render:
        env_config.update(render_video=True)
        env_config['camera_angle'] = args.camera_angle
    else:
        env_config.update(render_video=False)

    if args.use_gui:
        env_config.update(use_gui=True)
    else:
        env_config.update(use_gui=False)

    env_config.update(use_real_robot=args.use_real_robot)

    if args.store_actions:
        env_config['store_actions'] = True

    if args.log_obstacle_data:
        env_config['log_obstacle_data'] = True

    if args.save_trajectory_plot:
        env_config['save_trajectory_plot'] = True

    if args.plot_actual_torques:
        env_config['plot_actual_torques'] = True

    if args.use_thread_for_control_rate_sleep:
        env_config['use_thread_for_control_rate_sleep'] = args.use_thread_for_control_rate_sleep

    if args.obstacle_scene is not None:
        env_config['obstacle_scene'] = args.obstacle_scene

    if args.online_trajectory_duration is not None:
        env_config['online_trajectory_duration'] = args.online_trajectory_duration

    if args.plot_trajectory:

        env_config['plot_trajectory'] = True

        if args.plot_acc_limits:
            env_config['plot_acc_limits'] = True

        if args.plot_actual_values:
            env_config['plot_actual_values'] = True

        if args.plot_computed_actual_values:
            env_config['plot_computed_actual_values'] = True

        if args.plot_joint is not None:
            env_config['plot_joint'] = args.plot_joint

    if args.random_agent:
        env_config['random_agent'] = True

    if args.real_robot_debug_mode:
        env_config['real_robot_debug_mode'] = True

    if args.no_self_collision:
        env_config['no_self_collision'] = True

    if args.time_step_fraction_sleep_observation is not None:
        env_config['time_step_fraction_sleep_observation'] = args.time_step_fraction_sleep_observation

    if args.control_time_step is not None:
        env_config['control_time_step'] = args.control_time_step

    checkpoint_config['num_workers'] = 0
    checkpoint_config['env_config'] = env_config

    if args.seed is not None:
        checkpoint_config['seed'] = args.seed
        env_config['seed'] = args.seed

    args.config = checkpoint_config
    args.run = "PPO"
    tune.register_env(SafeMotionsEnv.__name__,
                      lambda config_args: SafeMotionsEnv(**config_args))
    args.env = checkpoint_config['env']
    args.out = None
    ray.init()
    cls = rollout.get_trainable_cls(args.run)
    agent = cls(env=args.env, config=args.config)
    agent.restore(checkpoint_path)
    if args.seed is not None:
        np.random.seed(args.seed)
    rollout_manually(agent, evaluation_dir)




# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import numpy as np
import logging
from abc import ABC
from safemotions.envs.safe_motions_base import SafeMotionsBase


def normalize_joint_values(values, joint_limits):
    return list(np.array(values) / np.array(joint_limits))


class TargetPointReachingReward(ABC, SafeMotionsBase):
    # optional action penalty
    ACTION_THRESHOLD = 0.9
    ACTION_MAX_PUNISHMENT = 1.0
    # braking trajectory collision penalty
    MIN_DISTANCE_MAX_THRESHOLD = 0.05  # meter; a lower distance will lead to punishments
    MIN_DISTANCE_MAX_PUNISHMENT = 1.0
    # braking trajectory torque penalty
    MAX_TORQUE_MIN_THRESHOLD = 0.9  # rel. abs. torque threshold
    MAX_TORQUE_MAX_PUNISHMENT = 1.0

    def __init__(self,
                 *vargs,
                 normalize_reward_to_frequency=False,
                 punish_action=False,
                 punish_braking_trajectory_min_distance=False,
                 punish_braking_trajectory_max_torque=False,
                 action_punishment_min_threshold=ACTION_THRESHOLD,
                 action_max_punishment=ACTION_MAX_PUNISHMENT,
                 braking_trajectory_min_distance_max_threshold=MIN_DISTANCE_MAX_THRESHOLD,
                 braking_trajectory_min_distance_max_punishment=MIN_DISTANCE_MAX_PUNISHMENT,
                 braking_trajectory_max_torque_min_threshold=MAX_TORQUE_MIN_THRESHOLD,
                 braking_trajectory_max_torque_max_punishment=MAX_TORQUE_MAX_PUNISHMENT,
                 target_point_reward_factor=1.0,
                 **kwargs):
        super().__init__(*vargs, **kwargs)

        # reward settings
        self.reward_range = [0, 1]
        self._normalize_reward_to_frequency = normalize_reward_to_frequency

        self._punish_action = punish_action
        self._action_punishment_min_threshold = action_punishment_min_threshold
        self._action_max_punishment = action_max_punishment

        self._punish_braking_trajectory_min_distance = punish_braking_trajectory_min_distance
        self._punish_braking_trajectory_max_torque = punish_braking_trajectory_max_torque
        self._min_distance_max_threshold = braking_trajectory_min_distance_max_threshold
        self._min_distance_max_punishment = braking_trajectory_min_distance_max_punishment
        self._max_torque_min_threshold = braking_trajectory_max_torque_min_threshold
        self._max_torque_max_punishment = braking_trajectory_max_torque_max_punishment

        self._target_point_reward_factor = target_point_reward_factor

    def _get_reward(self):
        info = {'average': {'reward': 0,  'action_punishment': 0, 'target_point_reward': 0,
                            'braking_trajectory_min_distance_punishment': 0,
                            'braking_trajectory_max_torque_punishment': 0},
                'max': {'reward': 0, 'action_punishment': 0, 'target_point_reward': 0,
                        'braking_trajectory_min_distance_punishment': 0,
                        'braking_trajectory_max_torque_punishment': 0},
                'min': {'reward': 0, 'action_punishment': 0, 'target_point_reward': 0,
                        'braking_trajectory_min_distance_punishment': 0,
                        'braking_trajectory_max_torque_punishment': 0
                        }}

        # action punishment

        reward = 0
        target_point_reward = 0
        action_punishment = 0
        braking_trajectory_min_distance_punishment = 0
        braking_trajectory_max_torque_punishment = 0

        if self._robot_scene.obstacle_wrapper.use_target_points:
            if self._punish_action:
                action_punishment = self._compute_action_punishment()

            target_point_reward = self._robot_scene.obstacle_wrapper.get_target_point_reward()

            if self._punish_braking_trajectory_min_distance or self._punish_braking_trajectory_max_torque:
                braking_trajectory_min_distance_punishment, braking_trajectory_max_torque_punishment = \
                    self._robot_scene.obstacle_wrapper.get_braking_trajectory_punishment(
                        minimum_distance_max_threshold=self._min_distance_max_threshold,
                        maximum_torque_min_threshold=self._max_torque_min_threshold)

            reward = target_point_reward * self._target_point_reward_factor \
                     - action_punishment * self._action_max_punishment

            if self._punish_braking_trajectory_min_distance:
                reward = reward - braking_trajectory_min_distance_punishment * self._min_distance_max_punishment

            if self._punish_braking_trajectory_max_torque:
                reward = reward - braking_trajectory_max_torque_punishment * self._max_torque_max_punishment

        if self._normalize_reward_to_frequency:
            # Baseline: 10 Hz
            reward = reward * self._trajectory_time_step / 0.1

        info['average'].update(reward=reward,
                               action_punishment=action_punishment,
                               target_point_reward=target_point_reward,
                               braking_trajectory_min_distance_punishment=braking_trajectory_min_distance_punishment,
                               braking_trajectory_max_torque_punishment=braking_trajectory_max_torque_punishment)

        info['max'].update(reward=reward,
                           action_punishment=action_punishment,
                           target_point_reward=target_point_reward,
                           braking_trajectory_min_distance_punishment=braking_trajectory_min_distance_punishment,
                           braking_trajectory_max_torque_punishment=braking_trajectory_max_torque_punishment
                           )

        info['min'].update(reward=reward,
                           action_punishment=action_punishment,
                           target_point_reward=target_point_reward,
                           braking_trajectory_min_distance_punishment=braking_trajectory_min_distance_punishment,
                           braking_trajectory_max_torque_punishment=braking_trajectory_max_torque_punishment
                           )

        # add information about the jerk as custom metric
        curr_joint_jerk = (np.array(self._get_generated_trajectory_point(-1, key='accelerations'))
                           - np.array(self._get_generated_trajectory_point(-2, key='accelerations'))) \
                          / self._trajectory_time_step

        curr_joint_jerk_rel = normalize_joint_values(curr_joint_jerk, self._robot_scene.max_jerk_linear_interpolation)
        jerk_violation = 0.0

        for j in range(self._num_manip_joints):
            info['average']['joint_{}_jerk'.format(j)] = curr_joint_jerk_rel[j]
            info['average']['joint_{}_jerk_abs'.format(j)] = abs(curr_joint_jerk_rel[j])
            info['max']['joint_{}_jerk'.format(j)] = curr_joint_jerk_rel[j]
            info['min']['joint_{}_jerk'.format(j)] = curr_joint_jerk_rel[j]

        max_normalized_jerk = np.max(np.abs(curr_joint_jerk_rel))
        if max_normalized_jerk > 1.002:
            jerk_violation = 1.0
            logging.warning("Jerk limit exceeded: %s", max_normalized_jerk)

        info['max']['joint_jerk_violation'] = jerk_violation

        return reward, info

    def _compute_action_punishment(self):
        # The aim of the action punishment is to avoid the action being too close to -1 or 1.
        action_abs = np.abs(self._last_action)
        max_action_abs = max(action_abs)
        action_punishment = (max_action_abs - self._action_punishment_min_threshold) / \
                            (1 - self._action_punishment_min_threshold)

        return max(min(action_punishment, 1), 0) ** 2

